#!/usr/bin/perl
# (c) napp-it  2012-2022 replicate

# 23. apr  18 inc. local repl -r, send_options, pv_limit
# last modification: monitor remote proc: end_proc, wait
# 2.10. check if pool is available, not only fs
# 22.6. check remote id in monitor
# 21.3.17 new keep values
# 2.5.17 monitor retry once
# 24.5.17 line 289, check local
# 21.dec keep/hold
# 28.12 snap month can be month: dec, Dez, Dec
# 7.7.18 BE: do not set ro
# 9.11.18 keep h on nov
# 21.1. linux repli (local, source and target)
# 16.5. ask remote via hostip, not host
# 17.7.19 rec_option (receive option)
# 23.10.19 netcat receive par -d makes problems on Proxmox6, tested without -d on ubunto and prpxmox
# 4.11.19 netcat on ProxMox requited a -w 10 on zfs receive
# 19.11.19 longer joblog

# 24.3.2020 keep days
# 2.4.2020 inherit unshare
# 15.5.2020 daisy chain replication
# 30.12.2020 keep days,monthy, year: max
# 02.03.2021 monitor timeout 600s
# 20.01.2022 keep month diff problem
# 08.feb 2022 nc on centos nc -l nr (debian nc -l -p nr)
# 10.feb 2022 rec transfer, set mountpoint to /pool/fs (or source is set), option writable=yes
# 22.mar 2022 keep month problem snap from current year
#             keep debug parameter D (do not destroy), ymdsh (details year,month,day,stunde,hold)
# 8.apr 2022 ncwait from about  > settings on sender side
# 13. apr 2023 ESXi snaps (pre/post)

# request replication job: receiver part only !!
# debug: run perl ..job-replicate.pl run_id_time manually
# ex                                 run_12345678 123  # 12345678=jobid, 123=debug mode

# howto:
# in the run part: start receiver
# in the monitor part: wait until receiver is ready, then start sender
# monitor sender and receiver

# use netcat on port 9090 like
# nc -b 32768 -d -l -p 9090 | zfs receive -F data/filesystem
# zfs send data/filesystem@1 data/filesystem@2 | nc -b 32768 -w 20 192.168.88.5 9090

# mbuffer on port 9090 like
# mbuffer -s 128k -m 1G -I 9090 | zfs receive -F data/filesystem
# zfs send -i data/filesystem@1 data/filesystem@2 | mbuffer -s 128k -m 1G -O 10.0.0.1:9090


#use strict;                                                   # disable for syntax check
use Fcntl qw(:flock);                                          # filelocking
$|=1;                                                          # keine Pufferung, no buffering

use vars qw($par $id $action $f $t @t $p @p %current $error $info %job $buf $os %sys $ZFS @keeph @keepd @keepm @keepy $keepfs $holdnr $debug $ncwait);

# netcat buffer size
$buf="131072";


 # common lib
 $f=`uname -a`;
 if ($f=~/^Linux/){
       $os="linux";
       $sys{'sbinpath'}='/sbin/';
       $ZFS="/sbin/zfs";
 } else {
       $os="illumos";
       $sys{'sbinpath'}='/usr/sbin/';
       $ZFS="/usr/sbin/zfs";
 }
 require "/var/web-gui/data/napp-it/zfsos/_lib/scripts/task-lib.pl";        #
 require "/var/web-gui/data/napp-it/zfsos/_lib/$os/grouplib.pl";
 require "/var/web-gui/data/wwwroot/cgi-bin/admin-lib.pl";


$par=$ARGV[0];
 if (!(defined($par))) {

     $ncwait=60;
     $t=`cat /var/web-gui/_log/napp-it.cfg | grep nc_wait`;      chomp($t);
     $t=~s/\D//gs;
     if ($t>10) { $ncwait=$t; }
     &log("job-replication |  missing parameter","m");
     print "job-replication 48: missing parameter ex run_1234567_12345\nncwait=$ncwait\n";
     exit;
 }

 #debug=012 etc
 $debug=$ARGV[1];                                  # ex 3=keep debug
 if ($debug > 10000) { $debug=""; }                # debug mode ex 123 or time()

 #par: action_job timestamp
 $par=~s/ .*//;
 chomp($par);

 # parameter: action_$job time
 if (!($par =~/\w+\_\d+$/)) {
     print "job-replication 57: wrong par $par\n";
     exit;
 }
 @t=();
 @t=split(/\_/,$par);    # action job
 $action=$t[0];
 $action=~s/[^a-z]//g;

 $id=$t[1];
 chomp($id);
 #############################################

if (-f "/tmp/nappit/job-replicate_$id.request")   {  `sudo rm /tmp/nappit/job-replicate_$id.request`;   }        # remove request ex /tmp/nappit/job-replicate_1350378156.request

# run?
if ($action eq "")                     {  &my_log_last("wrong action - must be not empty");     exit; }
if (!("run,cancel,delete"=~/$action/)) {  &my_log_last("wrong action - must be run: $action");  exit; }

#cancel or delete
if ($action eq "cancel") { &my_cancel; $info="cancel job"; &my_fast_end; exit; }
if ($action eq "delete") { &my_delete; exit; }                                       # call cancel, log, exit in sub

# get par -> %current
 &task_read_jobpar($id);
 &my_par($id);


if ($action eq "run") {

   if (-f "/tmp/nappit/$id.cancel")    {  `sudo rm /tmp/nappit/$id.cancel`;  }

   # do not start this job replicate.pl run_123 > twice (fork)
     $t=`ps axw | grep run_$id | grep -v grep | grep -v color`;  # color_ linux
     @t=();
     @t=split(/\n/,$t);

     # job + fork: 2 allowed
     if (@t > 2) {
         &my_log_id("cancel next, job is running");
         &my_log_last("next replication cancelled<br>job is running");
         exit;
     } else {
         &my_log_id("next replication started");

         # optionally delete incomplete last
         if ($job{'port'} > 49000) {
            `pkill -f 'p $job{'port'}'`;
         }

     }
}


# optionally delete

# log par
  $t=&print_my_par;  $t=~s/\n/<br>/gs;
  &my_log_last("main line 90<br>$t");

#start detailed log
 &task_gettime;
 $t="/var/web-gui/_log/jobs/$id.last";
    open (PF, ">$t");
      flock(PF, LOCK_EX);
      print PF "$current{'timestamp'}\tmain line 76\tstart job-replicate with parameter:<br>id=$id, action=$action, par=$par\n";
    close (PF);

# logfile receiver started
 if (-f "/tmp/$id.receiver") { `rm /tmp/$id.receiver`; }

 # opt pre job
 if (-f "/var/web-gui/_log/jobs/$id.pre") {
       &my_log_last("() execute shell script $id.pre");
       $t=`sudo sh /var/web-gui/_log/jobs/$id.pre`;
       &my_log_last("result of pre script\: $t");
 }

  # ESXi snaps
  if ($job{'esxi_ip'} =~/\d/) {
    $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl create_$id`;   # create ESXi snaps
  }

 # start replication
 &my_run;
 exit;


################# monitor job (parent process): wait until receiver is ready, opt request remote sender and monitor
 sub my_monitor {
#################
my ($r,$p,$ok,@p,$s,$t,$l1,$l2,$r1,$r2,$l,$monitor,$wait,@t,$retry);

    # debug: wait = disable monitor
    #wait; exit;

    $ok="1";
    $retry="0";

    sleep(3);
    while ($ok) {
            if (-f "/tmp/$id.receiver") {
                 $ok="0";
                 last;
            }
            sleep(10);
            print "142 monitor waits for receiver\n";
    }

    my $remote="0";
    $wait="0";
    #&my_log_last("start my_monitor, parent fork<br>host $job{'host'} >> myhost $job{'myhost'}");

    while (1) {

        $ok="0";
        $l1="";                                              # local ok?   send or nc
        $r="";                                               # remote ok=? nc receive
        $error="";

        $l=`ps axw | grep web-gui`;                          # all process

        if ($job{'host'} eq "$job{'myhost'}") {              # local replication

               @p=();
               @p=split(/\n/,$l);
               foreach $t (@p) {
                    if ($t=~/grep/)   { next; }
                    if (!($t=~/zfs/)) { next; }
                    if ($t=~/zfs receive .*$job{'destination'}/) { $l1="1"; }   # zfs receive .*test3/wd2tb/a
                    if ($t=~/zfs send .*$id/)                    { $l2="1"; }   # zfs send .*1353434134

                    &my_log_last("monitor, local zfs send or receive process: $t");
               }

               if (($l1 eq "") || ($l2 eq "")) {
                     # job finished with or without error
                     &my_log_last("monitor, no local zfs send or receive process");
                     wait;                              # wait, exit by parent/receiver
               }
                $t=`$ZFS list -H -o used $job{'destination'} 2>&1`;
                chomp ($t);
                if (($l1 ne "1") && ($l2 ne "1"))  {
                     if ($t=~/dataset does not exist/s)  {                  # not created and repli not running
                          &my_log_last("zfs list -H -o used $job{'destination'}\n$t");
                          $info="error, could not create target ZFS $job{'destination'} (ZFS or pool problem)";

                          &my_log_id("error","$info");
                          wait;                         # wait, exit by parent/receiver
                     }
               }

        } else {  # remote

               $l1="0";                                  # local nc not ok
               $r1="0";                                  # remote send not ok
               $remote="0";

               #check local
               @p=();
               @p=split(/\n/,$l);
               foreach my $t (@p) {
                    if ($t=~/grep/) { next; }
                    if ($t=~/nc .*$job{'port'}/)   {    #/var/web-gui/data/tools/nc/nc -b 131072 -d -l -p 507219
                         $l1="1";                       # lokal ok
                    }
                    # option
                    # if ($t=~/zfs receive .*$job{'destination'}/) { $l1="1"; }   # zfs receive .. backup3/nfs2 ##   output too short !!
               }

               #check remote
               $p=&grouplib_ask_remote_pslist($job{'hostip'},,"job-replicate");        # proc;;proc;;

               if ($p=~/key is wrong/s) {
                    &my_log_last("end monitor, error, re-add host to appliance group: $p");
                    &my_log_id("ask_remote_pslist $job{'hostip'}","key is wrong, re-add host to appliance group");
               }

               #timeout?
               if ($p=~/error|timeout/s) {
                    $retry++;
                    if ($retry <=1 ) {
                         &my_log_last("Monitor: Remote control error or timeout, retry once:\n$p");
                         sleep(30);
                         next;                 # try once again, ex short outage
                    }
                    &my_log_last("end monitor, error: $p");
                    $info="check network, $p";

                    &my_log_id("error","$info");
                    sleep(20);                                 # was 10: 2.3.21
                    &my_fast_end;
               } else {
                    $retry="0";
               }

               @p=();
               @p=split(/;+/,$p);
               $p="";
               foreach my $t (@p) {
                    if ($t=~/grep/)                     { next; }
                    if (!($t=~/nc |zfs /))              { next; }

                    #$p.=" -> $t\n";

                    # optional
                    # nc .*192.168.88.23 51089 -> /var/web-gui/data/tools/nc/nc -b 131072 -d -l -p 50721
                    #if ($t=~/nc .*$job{'port'}/)        {           # can be a problem with long strings
                    #    $r1="1";
                    #    $p.="$t\n";
                    #}

                    # zfs send .*1348240568  ex /usr/sbin/zfs send tank/nfs-vw1@1466590204_repli_zfs_backup1_nr_1
                    if ($t=~/zfs send .*$id/)           {           # send id is relevant for remote
                        $r1="1";              # remote send ok
                        $remote="1";          # remote ok
                        $wait="0";            # no extra wait
                        $p.="$t\n";
                    }
               }




               # log remote once
               if ($monitor ne "1") {
                    if (!($p=~/\w/s)) { $p="remote nc and zfs send not running or finished, wait for postprocessing";}
                    &my_log_last("Monitor: Remote proc:\n$p");

                    # optional comment to log remote always
                    $monitor="1";
               }

               if ($remote eq "0") {                 # remote not answering or finished
                  # local and remote terminated
                    if (($l1 eq "0") && ($wait eq "1")) {               # receiver running, sender not
                            $error="";
                            &my_log_last("monitor info, replication terminated: receive=$l1, remote send=$r1");
                            exit;
                    } else {

                          if ($wait eq "0") {
                              $wait="1";
                              sleep(900);             # long wait ex for postreplication with many snaps, kill on error only
                              next;
                          } else {

                              # check if dest_snap was created
                              print "test $current{'next_dest_snap'}\n";
                              if ($current{'next_dest_snap'}=~/\@/) {
                                  $t=`$ZFS list -H -o name -t snapshot | grep $current{'next_dest_snap'}`;
                                  chomp ($t);
                                  if ($t=~/\@/) {
                                       &my_log_last("monitor info, next dest_snap ok: $current{'next_dest_snap'}");
                                       &my_log_id("monitor info","next dest_snap ok: $current{'next_dest_snap'}");
                                       print "monitor 330: next snap ok: $current{'next_dest_snap'}\n";
                                       &my_end_proc;
                                       exit;
                                  }  else {
                                     &my_log_last("monitor info, next dest_snap error: $current{'next_dest_snap'}");
                                  }
                              }

                              sleep(5);             # extra wait ex for postreplication with many snaps
                              $error="monitor info, replication terminated (receiver running but sender is not running)";
                              &my_log_id("error, monitor info","replication terminated: local receive=$l1, remote send=$r1 - check zpool status");
                              &my_log_last("error, monitor info, replication terminated: local receive=$l1, remote send=$r1 - check zpool status");
                              &my_end_proc;
                              # error if target snap not created

                              #last action?
                              exit;
                          }
                    }
               }

               $t=`$ZFS list -H -o used $job{'destination'} 2>&1`;
               chomp($t);

               if (($l1 eq "0") || ($r1 eq "0"))  {                         # transfer not ok
                     if ($t=~/dataset does not exist/s)  {                  # not created and repli not running
                          &my_log_last("zfs list -H -o used $job{'destination'}\n$t");
                          $info="error, could not create target ZFS $job{'destination'} (Remote appliance, pool, stream or network problem)";
                          $error="error my_monitor";
                          &my_log_id("error","$info");
                     }
                     wait;  # wait, exit by parent/receiver
               }
        }

        #progress in file /tmp/nappit/id.progress -> sourcesize in MB - %
        #print "job-repli 151 update progress\n";

        if (($t > 0) && ($current{'snapsize'} eq "")) {   # initial repli
           if (-f "/tmp/nappit/$id.progress") {           # source_target_percent
               $r=`cat /tmp/nappit/$id.progress`;
               $r=~s/\_.*//s; $r=~s/\n.*//s;              # source size in MB

               $t=&my_val2mb($t);
               $p=$t/$r;
               $p=sprintf ("%.2f", $p); $p=$p*100;
               if ($p=~/0$/) {
                 &my_log_last("initial send progress for parent/base fs (log every 10%) $p %");
               }  else {
                 # &my_log_last("initial send progress (log every 10%) data! $r\_$t\_$p");       # optionally
               }


               $p="$r\_$t\_$p";
               `echo $p > /tmp/nappit/$id.progress`;
           }
        }

        &my_wait_and_check(10);     # check cancel, wait 10s
    } # /while
 }





######## run job (child process)
sub my_run {
###########

    # check if destination exists (initial)


#test
    # `zfs destroy -f -r test3/test2`;
    #&my_end_proc;    # just to be sure

    # check if pool is mounted
    my $r=$job{'destination'};  # pool/fs
    $r=~s/\/.*//;               # pool
    $r = `$ZFS list -H -o name \'$r\' 2>&1`;
    chomp $r;

    if ($r=~/not exist/)  {
        $info="pool $r not mounted";
        &my_log_last("$info");
        &my_log_id("error","$info");
        &my_cancel; $error="error"; $info="pool $r not mounted";  &my_fast_end; exit;
    }

    $r = `$ZFS list -o name | grep \'$job{'destination'}\'`;

    chomp($r);
    @t=split(/\n/,$r);
    $r="";
    foreach my $t (@t) {
        if ($t eq $job{'destination'}) { $r="$job{'destination'}"; }      # to avoid substr errors
    }

    #debug
    #print "$job{'host'}\n$job{'hostip'}\n"; exit;

    ################
    if ($r eq "") {  # initial
           # opt 1: local transfer (host=myhost);
           if ($job{'hostip'} eq "127.0.0.1") {
              &log("$id <br>job-replicate 243 | start replication<br>local initial | $job{'host'} -> $job{'myhost'}<br>$job{'source'} -> $job{'destination'} | ","m");     #monitor
              &my_local_initial_replication;
           } else {
              &log("$id <br>job-replicate 2246 | start replication<br>remote initial | $job{'host'} -> $job{'myhost'}<br>$job{'source'} -> $job{'destination'} | ","m");     #monitor
             &my_remote_initial_replication;
           }
           if (-f '/tmp/nappit/zfs.cfg') { unlink('/tmp/nappit/zfs.cfg'); }
    } else {          # delta
           if ($job{'hostip'} eq "127.0.0.1") {
              &log("$id <br>job-replicate 252 | start replication<br>local incremental | $job{'host'} -> $job{'myhost'}<br>$job{'source'} -> $job{'destination'} | ","m");     #monitor
              &my_local_delta_replication;
           } else {
              &log("$id <br>job-replicate 255 | start replication<br>remote incremental | $job{'host'} -> $job{'myhost'}<br>$job{'source'} -> $job{'destination'} | ","m");     #monitor
              &my_remote_delta_replication;
           }
    }

     if (-f "/tmp/nappit/$id.progress")  {  unlink("/tmp/nappit/$id.progress"); }
     if (-f "/tmp/nappit/$id.cancel")    {  unlink("/tmp/nappit/$id.cancel");   }

     &my_log(); # also in sub
     exit;
 }


#################################
sub my_remote_initial_replication {
#################################

     &my_log_last("start remote initial replication");

     my ($r,@v,$v,$cmd,$timer,$ok,$var,$p1,$p2,$nc,$cmd,$t);

    #  ex -p 53522
     # do not start this job replicate.pl run_123 twice
     $t=`ps axw | grep "p job{'port'}" | grep -v grep | grep -v color`;  # color_ linux
     chomp($t);

     # do not run twice
     if ($t ne "") {
         &my_log_id("cancel next, receive is running");
         &my_log_last("next replication cancelled<br>receive is running");
         exit;
     }

     $ncwait=60;
     $t=`cat /var/web-gui/_log/napp-it.cfg | grep nc_wait`;    chomp($t);
     $t=~s/\D//gs;
     if ($t>10) { $ncwait=$t; }
     &my_log_last("ncwait in remote call: $ncwait");

     # opt delete old snap with this id (restart)
     &my_del_snap($id);          #del locally

     &log("$id <br>job-replicate 277  | request remote zfs | ","m");           # monitor

     #size of source
     $r=&grouplib_ask_remote_zfslist("$job{'hostip'}","","","job-repli 280");    # commalist


     #source fs not available?? lines separated by ,
     $t=" $r"; $t=~s/,/ /gs; $t=" $t ";
     if (!($t=~/ $job{'source'} /s)) {
        &my_log_last("source zfs missing: $job{'source'}");
        $error="error, source filesystem $job{'source'} not found";  &my_fast_end;  exit;
     }

     #timeout?
     if ($r=~/error|timeout|offline/s) {
          &my_log_last("ask zfslist: $error");
          $info="check network, $r"; $error="error"; &my_fast_end;  exit;
     }

     #get my ip
     $job{'myip'}=&grouplib_ask_remote_get_my_ip($job{'hostip'});
     if (!($job{'myip'}=~/^\d+\.\d+/)) {  $info="request my ip 287 ($job{'hostip'} -> $job{'myip'})"; $error="error";  &my_fast_end;  exit;  }


     my @rzfs=split(/,/,$r);
     my $size;
     foreach my $s (@rzfs) {
        if (!($s=~/^$job{'source'} +/))  { next; }
        $s=~s/ +/\t/g;
        @v=split(/\t/,$s);                                                      #name used avail refer mount

        # job zfs-source size needed in monitor
        $size=$v[1];
        $r=&my_val2mb($v[1]);
        $r="$r\_";
       `echo $r > /tmp/nappit/$id.progress`;
     }


     #check remote snaps
     $r = &grouplib_ask_remote_snaplist("$job{'hostip'}");
     #timeout?
     if ($r=~/error|timeout/s) {
          &my_log_last("ask remote snaps: $r");
          $info="check network, $r"; $error="error";  &my_fast_end; exit;
     }


     my @rsnaps=();
     @rsnaps=split(/,/,$r);                                                             # commalist !!

     # opt delete all source snaps in case of restart
     foreach my $s (@rsnaps) {
           if (!($s=~/^$job{'$source'}.*\@$id\_repli/)) { next; }                       # inkl recursiv snaps
           $t=&grouplib_ask_remote_destroy_snap("$job{'hostip'}","snap_name=$s");
     }

     #start receiver
     # fork parent=receiver, parent=send
     my $child=fork();

     if (!($child)) {       # child process=receiver

          if ($current{'receiver_ok'} eq "1") {    wait; }  # do not restart
           my $t1=time();
#print __line__.$os;
           if ($os eq "linux") {
                      $nc="nc";

                      $cmd="$nc -w 60 -l -p $job{'port'} | $ZFS receive -Fv $job{'destination'} 2>&1"; # 4.11.19, ProxMox, Debian
                      $r=`nc 2>&1`;
                      if ($r=~/\-p source_port/s)   {
                         $cmd="$nc -w 60 -l $job{'port'} | $ZFS receive -Fv $job{'destination'} 2>&1"; # 8.feb 2022, centos
                      }


           } else {
                      # no wait in listen mode
                      `/usr/bin/chmod 755 /var/web-gui/data/tools/nc/nc`;
                      if ($buf > 0) {
                         $nc="/var/web-gui/data/tools/nc/nc -b $buf";
                      } else {
                        $nc="/usr/bin/nc";
                      }

                      $cmd="$nc -d -l -p $job{'port'} | $ZFS receive $job{'rec_option'} -Fv $job{'destination'} 2>&1";
                      $cmd=~s/ +/ /g;
           }
           &my_log_id("start receive","$cmd");
           &my_log_last("start receive: $cmd");
           &log("$id <br>job-replicate 347 | start receive | $cmd |  ","m");               # monitor

          `touch /tmp/$id.receiver`;        # signal for monitor

           $current{'receiver_ok'} = "1";
           $r=`$cmd`;
           chomp ($r);

           # error or info?
           if ($r ne "") {
              $r=~s/\s+/ /gs;
              if ($r=~/receiving/) {
                  $r=~s/into/into<br>/;
                  $r=~s/received/<br>received/;
                  &my_log_last("receiver message\n$r");
              } else {
                  &my_log_last("error, receiver message\n$r");
              }

             # error or info?

             # failed to read from stream can be an info
             if ($r=~/failed to read from stream/s) {
                &my_log_id("receive finished with warning 'failed to read from stream', <br>can be a message only: check if target snap is created");
                $r="";
             }

             # cannot mount can be an info
             if ($r=~/cannot mount/s) {
                &my_log_id("receive finished with cannot mount, <br>warning: check if target snap is created","$r");
                $r="";
             }

             if ($r eq "") {
                 &my_log_id("receiver message\n$r");
             }
           }

           chomp($r);
           my $t2=time();
           $t=$t2-$t1;

           &log("$id <br>job-replicate| receive finished after $t s| $r ","m");       # monitor
           &my_log_id("receive finished","time $t s <br>$r");
           &my_log_last("end receiver after $t s");

           &my_postreplication;

           ## check valid transfers
           my $next_dest_snap="$job{'destination'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1";   # backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266


          # check for sucessfully created snap
          #read new snaps
          my $t=`$ZFS list -H -r -o name -t snapshot`;
          chomp($t);
          my @s=split(/\n/,$t);

          my $ok="0";
          foreach my $s (@s) {
               if ($s eq "$next_dest_snap") {             #ok
                   $ok="1";
               }
          }

          if ($ok eq "0") {  # error, no target, del source snap
                   $error="missing destination snap: $next_dest_snap";
                   &my_log_last("missing destination snap: $next_dest_snap");
                   &log("$id error<br>job-replicate | destination snap not created |  | missing<br>$next_dest_snap ","m");     #monitor
                   $t=&grouplib_ask_remote_destroy_snap("$job{'hostip'}","snap_name=$job{'source'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1");
          }

          $t=$t2-$t1;
          my $used=`$ZFS list -H -o used $job{'destination'}`;
          chomp($used);
          $used=&my_val2mb($used);

          my $perf;
          if ($t>0) {
            $perf=$used/$t;
            $perf=sprintf ("%.1f", $perf);
          }

          if (!($used > 0)) { $error="error"; }
          $info="initial remote replication finished (time: $t s, size: $used MB, performance: $perf MB/s)"; $info=~s/ +/ /g;

          &my_log_last("$info");
          &my_log_jobs;

          &my_fast_end;
          wait; exit;

     ###############################################
     } else {      # parent, wait for receiver, request send

           if ($current{'send_ok'} eq "1") { wait; }     # do not start twice

           $timer=time() +60;           #max wait
           $ok="0";
           $p1=""; $p2="";
           do {
               $r=`ps axw`;
               @v=();
               @v=split(/\n/,$r);
               foreach $v (@v) {
                  if ($v=~/grep/)   { next; }
                  if ($v=~/nc .*$job{'port'}/)                      { $p1="1"; }
                  if ($v=~/zfs receive -F $job{'destination'}/)     { $p2="1"; }
               }
               if (($p1 eq "1") && ($p1 eq "1")) { $ok="1"; } else {  sleep(2); }

           } until (($ok) or (time() >= $timer));
           if (time() >= $timer) {
               &my_log_last("error, timeout receiver");
               &my_log_id("error, timeout receiver");
               $error="error";  $info="joblib, receiver failed, timeout error";  &my_fast_end; exit;
           }


          $var ="id=$id\n";

          #var for remote send  ,wait set on sender side
          if ($os ne "linux") {
             $t="/usr/bin/nc $job{'myip'} $job{'port'}";       # sender parameter, options nc, bds-nc, mbuffer
          } else {
             $t="nc $job{'myip'} $job{'port'}";               # sender parameter (overwrite on sender!), options nc, bds-nc, mbuffer
          }
          $var.="src_interface=$t\n";                         # sender par napp-it <= 0.8

          $var.="snap_name1=\n";
          $var.="snap_name2=$job{'source'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1\n";
          $var.="snap_rec=$job{'snap_rec'}\n";
          $var.="send_rec=$job{'send_rec'}\n";
          $var.="send_dedup=$job{'send_dedup'}\n";
          $var.="buffer=$buf\n";

          $var.="send_option=$job{'send_option'}\n";
          $var.="pv_limit=$job{'pv_limit'}\n";

          &my_log_id("initial send","$job{'host'}\: zfs send $job{'send_option'} $job{'send_rec'} $job{'source'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1 | $t");
          &log("$id <br>job-replicate 442 | request send | $job{'host'} ($size)\: zfs send $job{'send_option'} $job{'send_rec'} $job{'source'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1 | $var","m");     #monitor

          &my_log_last("request remote send \n$job{'host'}\n$var\n");

          $r = &grouplib_ask_remote_zfs_send ("$job{'hostip'}", $var,"","");

          # send error
          if ($r=~/error/) {
              $error="error, request send failed";  $info="joblib, request send failed: $r";  &my_fast_end; exit;
          }

          $current{'send_ok'}="1";

          &my_log_last("result\n$r");
          &my_monitor;                                              # monitor receiver

          wait;
          #exit;
     }

}

##############         rename snap?
sub my_daisy {
##############
  return; # todo
  my ($s,$r);
  my $snap=$_[0];

  if ($debug=~/1/) {
       print "\n726 daisy process $job{'daisy'} with $snap\n";
  }

  # daisy chain?, rename n-3 snap
  # zfs rename tank/home/cindys@083006 tank/home/cindys@today
  ###############################################
  if ($job{'daisy'}=~/\d+/) {
       $s=$snap; $s=~s/_nr_/_dc_nr_/;
       $s=~s/\@.+_repli/\@$job{'daisy'}_repli/;      # dc job id

       $r=`$ZFS list -H -o name $s 2>/dev/null`;
       if (!($r=~/\@/)) {   # rename snap
             #snap="$job{'source'}\@$id\_repli_zfs_$job{'myhost'}\__dc_nr_nn";
              $r="$ZFS rename $snap $s 2>&1"; $r =~s/ +/ /g;
                   $r=`$r`;
                   if ($debug=~/1/) {
                     print "\n738 rename daisy chain snap: $snap -> $s\n";
                   }
                   &my_log_last("935 rename daisy chain snap\n$s");
                   &my_log_id("rename daisy chain snap","$s");
               }
               return(1);
   }
   #/daisy
   return(0);
}

#################################
sub my_remote_delta_replication {
#################################

    my ($r,@v,$v,$cmd,$timer,$ok,$var,$p1,$p2,$nc,@s,$s,$dnr,$dnr2,$dnr3,$t,$used);

    #  ex -p 53522
     # do not start this job replicate.pl run_123 twice
     $t=`ps axw | grep "p job{'port'}" | grep -v grep | grep -v color`;  # color_ linux
     chomp($t);

     # do not run twice
     if ($t ne "") {
         &my_log_id("cancel next, receive is running");
         &my_log_last("next replication cancelled<br>receive is running");
         exit;
     }

     $ncwait=60;
     $t=`cat /var/web-gui/_log/napp-it.cfg | grep nc_wait`;    chomp($t);
     $t=~s/\D//gs;
     if ($t>10) { $ncwait=$t; }
     &my_log_last("ncwait in remote call: $ncwait");

     &my_log_last("start remote incremental replication: read snaps");

     #step 1 check for newest source dest snaps
     my $newest_src="0";
     my $newest_dest="0";

     my $newest_src_snap="";
     my $newest_dest_snap="";
     my $next_src_snap="";
     my $next_dest_snap="";

     #size of source
     $r=&grouplib_ask_remote_zfslist("$job{'hostip'}","","","job-repli 485");    #commalist

     #timeout?
     if ($r=~/error|timeout|offline/s) {
          &my_log_last("ask remote zfslist\n$r");
          $info="614, check network, $r"; $error="error"; &my_fast_end; exit;
     }

     #source fs not available??, lines separated by ,
     $t=" $r ";  $t=~s/,/ /gs;
     if (!($t=~/ $job{'source'} /s)) {
        &my_log_last("error source missing: $job{'source'}");
        $error="error, source filesystem $job{'source'} not found";   &my_fast_end; exit;
     }

     $r = &grouplib_ask_remote_snaplist("$job{'hostip'}");   #commalist
     &log("$id <br>job-replicate 449  | request remote snaps | $r","m");            #monitor

     #timeout?
     if ($r=~/error|timeout/s) {
          &my_log_last("error: $r");
          $info="check for network problems or timeout due to too many snaps (1min timeout), $r"; $error="error"; &my_fast_end;  exit;
     }

     @s=split(/,/,$r);

     ($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap) =&my_newest_snap_pair(@s);  # remote transfer, add remote zfs

     $current{'next_dest_snap'}=$next_dest_snap;
     # last        = last dest prior repli run
     # newest_dest = current base snap
     # next_dest   = new snap after repli run

     # check remote snaps for current nr
     $ok="0";
     foreach $s (@s) {
          chomp($s);
          $s=~s/ .*//;  # remove other info but name
          if ($s eq "$newest_src_snap") {
              $ok="1";
          }
     }
     if ($ok eq "0") {
          &my_log_last("error: needed source snap missing: $newest_src_snap");
          $info="job-repli, needed source snap missing: $newest_src_snap"; $error="error"; &my_fast_end; exit;  }


     #####
     # check remote snaps and opt delete next src snap (kept from unsuccessfull repli) or other than current
     $dnr="";
     $dnr2="";
     $dnr3="";
     if ($newest_dest_snap=~/_nr_\d+/) {
                   $dnr=$newest_dest_snap;
                   $dnr=~s/.*_nr_//;
                   $dnr2=$dnr+1;                                       # next (new) snapnr
                   $dnr3=$dnr-1;                                       # previous
     }
     if (($newest_src_snap ne "")  && ($id >0) && ($dnr >0))     {
            foreach my $s (@s) {
                       chomp($s);
                       if (!($s=~/$id\_repli/))     { next; }               # not my id

                       $s=~s/ .*//;  # remove other info but name

                       if ($s=~/_nr_$dnr2$/)          { next; }        # next nr   (keeps 3 snaps, comment for one snap)
                       if ($s=~/_nr_$dnr$/)           { next; }        # current  nr
                       if ($s=~/_nr_$dnr3$/)          { next; }        # previous  nr        # needed to destroy newest target snap and redo
                       &my_log_id("request remote snap destroy","$s");
                       $r=&grouplib_ask_remote_destroy_snap("$job{'hostip'}","snap_name=$s");
            }

            $current{'newest_dest_snap'}=$newest_dest_snap;      # needed for keep
            $t=$newest_dest_snap;
            $t=~s/.*\_//;                 # keeps nr
            if ($t>1) {
                $current{'last_dest_snap'}=$newest_dest_snap;
                $current{'last_dest_snap'}=~s/\d+$//;
                $current{'last_dest_snap'}.=$t;
            }

            # do the same with locals snaps

                $r=`$ZFS list -H -r -o name -t snapshot -s creation`;     # newest last

                chomp($r);
                @s=();
                @s=split(/\n/,$r);

                $t=$id."_recursiv";
                $t=$current{$t};                                             # yes
                # keep and hold does not work with recursive replication !!

                foreach my $s (@s) {
                              if (!($s=~/$id\_repli/))      { next; }        # not my snaps
                              if ($s=~/_nr_$dnr$/)          { next; }        # current  nr
                              if ($s=~/_nr_$dnr2$/)         { next; }        # next nr       (keep 3 snaps, comment for one snap)
                              if ($s=~/_nr_$dnr3$/)         { next; }        # previous nr

                              if ($s eq $last_dest_snap)    { next; }        # -2
                              if ($s eq $newest_dest_snap)  { next; }        # -1
                              if ($s eq $next_dest_snap)    { next; }        # newest

                              if ($dnr < 2) {
                                if ($s=~/\_1$|\_2$/)        { next; }        # keep on init
                              }

                              #daisy rename
                              $r=&my_daisy($s);
                              if ($r) {next; }

                              #keep snap?
                              if (&my_keep($s,$dnr)) {
                                   #print "debug 789: skip $s\n";
                                   next;
                              }

                              &my_log_id("zfs destroy","$s");

                              if ($debug=~/\w/) { print "xxx debug 793:  -- destroy $s\n"; }
                              if ($debug=~/D/) {         # Dry run
                                print "xxx debug 929:  -- would destroy $s\n";
                              } else {
                                `$ZFS destroy -f $s`;
                              }
               }
               #print "debug 802: end of snap processing\n";

     }
     ####

     #get my ip
     $job{'myip'}=&grouplib_ask_remote_get_my_ip($job{'hostip'});                # ip or name as fallback

     #timeout?
     if ($job{'myip'}=~/error|timeout/s) {
          &my_log_last("network error, $job{'myip'}");
          $info="job repli: network error, $job{'myip'}"; $error="error"; &my_fast_end;  exit;
     }

     if (!($job{'myip'}=~/^\d+\.\d+/))   {
         &my_log_last("network error, $job{'myip'}");
         $info="request my ip ($job{'hostip'} -> $job{'myip'})"; $error="error"; &my_fast_end;  exit;
     }


     #start receiver
     # fork parent=receiver, child=req send;  #my parent=$;


     ############################################### do #########################################
     my $child=fork();
     if ($child) {       # parent process=receiver

           #&my_log_last("init receiver fork");
           if ($current{'receiver_ok'} eq "1") { &my_log_last("double receive");   exit; }  # do not restart

           my $t1=time();

           if ($os eq "linux") {
                      $nc="nc";
                      $cmd="$nc -w 60 -l -p $job{'port'} | $ZFS receive -Fv $job{'destination'} 2>&1";   # 4.11.19, ProxMox, Debian
                      $r=`nc 2>&1`;
                      if ($r=~/\-p source_port/s)   {
                         $cmd="$nc -w 60 -l $job{'port'} | $ZFS receive -Fv $job{'destination'} 2>&1"; # 8.feb 2022, centos
                      }
           } else {
                      # no wait time in listen mode
                      `/usr/bin/chmod 755 /var/web-gui/data/tools/nc/nc`;
                      if ($buf > 0) {
                         $nc="/var/web-gui/data/tools/nc/nc -b $buf";
                      } else {
                        $nc="/usr/bin/nc";
                      }
                      # add process viewer pv
                      $cmd="$nc -d -l -p $job{'port'} | $ZFS receive $job{'rec_option'} -Fv $job{'destination'} 2>&1";
                      $cmd=~s/ +/ /g;
           }
           &my_log_id("start receiver","$cmd");
           #&log("$id <br>job-replicate  | start receive |","m");                       #monitor

           &my_log_last("start receiver \n$cmd");

           `touch /tmp/$id.receiver`;        # signal for monitor
           $current{'receiver_ok'}="1";


           ########## run receiver #####
           if ($debug=~/1/) { print "\nstart receiver $cmd\n"; }
           $r=`$cmd`;
           if ($debug=~/1/) { print "\nend receiver with $r\n"; }

           # error or info?
           if ($r ne "") {

              $r=~s/\s+/ /gs;
              if ($r=~/receiving/) {
                  $r=~s/into/into<br>/;
                  $r=~s/received/<br>received/;
                  &my_log_last("receiver message\n$r");
                  &my_log_id("receiver message","$r");
              } else {
                  &my_log_last("error, receiver message\n$r");
                  &my_log_id("receiver message","$r");
              }

             # error or info?

             # failed to read from stream can be an info
             if ($r=~/failed to read from stream/s) {
                &my_log_id("receive finished with warning 'failed to read from stream', <br>can be a message only: check if target snap is created");
                $r="";
             }

             # cannot mount can be an info
             if ($r=~/cannot mount/s) {
                &my_log_id("receive finished with cannot mount, <br>warning: check if target snap is created","$r");
                $r="";
             }

             if ($r eq "") {
                 &my_log_id("receiver message\n$r");
             }
           }

           chomp($r);
           my $t2=time();

           $t=$t2-$t1;
           $t.=" s ";

           &log("$id <br>job-replicate 630  | end receiver after $t| $r","m");        #monitor
           &my_log_id("end receive","zfs receive $id running $t <br>$r");
           &my_log_last("end receive $t");

           #########
           ## ok? ##
           #########

           #sleep(1);
           &my_postreplication;

           ## check valid transfers
           #my $next_dest_snap="$next_dest_snap";             #backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266

          # check for sucessfully created snap
          #read new snaps

          my $ok="0";
          my $snap="";

          if ($next_dest_snap=~/\@/) {
            sleep(2);
            $t=`$ZFS list -H -o name $next_dest_snap`;
            chomp($t);
          } else {
            $t="";
          }

          if ($t =~/\@/) {
               $ok="1";
               $snap=$next_dest_snap;
          }

          if ($ok eq "0") {  # error, no target, del source snap
                   $error="error check network, systemlog, poolstate, capacity, timeouts<br>on recursive transfers optionally delete newest target snap and retry with the former";
                   # &my_log_id("request remote snap destroy due dest snap missing","$next_src_snap");  # can give problems after a false detected  replication error
                   # $t=&grouplib_ask_remote_destroy_snap("$job{'hostip'}","snap_name=$next_src_snap"); # automatically destroyed on next run
                   $info="887 next destination snap $next_dest_snap from $job{'host'} was not created<br><br>";
          } else {
                   #
                   $t=$t2-$t1;

                   my $nr=$next_dest_snap; $nr=~s/.*nr\_//;
                   my $snap_r=`$ZFS list -H -o refer $snap`;
                   chomp($snap);

                   $info="ok incremental remote replication nr $nr finished (time: $t s) <br>new snap $snap, refer: $snap_r";
                   $info=~s/ +/ /g;
          }

          &my_log_last("$info");
          &my_log_jobs;

          &my_fast_end;
          exit;

     ###############################################
     } else {      # child, wait for receiver, request send

          #&my_log_last("init sender + monitor fork");
          if ($current{'send_ok'} eq "1") { &my_log_id("double send"); exit; }     # do not start twice

          sleep(2);   # wait 2s
           #$timer=time() + 5;                         # max wait
           #$ok="0";
           #$p1=""; $p2="";
           #sleep(1);
           #do {
           #    $r=`ps axw`;

               #print "1105 $r\n";
           #    @v=();
           #    @v=split(/\n/,$r);
           #    foreach $v (@v) {
           #       if ($v=~/grep/)   { next; }
           #       if ($v=~/nc .*$job{'port'}/)                      { $p1="1"; }
           #       if ($v=~/zfs receive -F $job{'destination'}/)     { $p2="1"; }
           #    }



           #    if (($p1 eq "1") && ($p2 eq "1")) { $ok="1"; } else {  sleep(2); }

           #} until (($ok) || (time() >= $timer));
           #if (time() >= $timer) {
           #    &my_log_last("error, timeout receiver");
           #    &my_log_id("error, timeout receiver");
           #    $error="error timeout ";
           #    $info="joblib, receiver failed, timeout"; &my_log_last("start receiver timeout"); &my_fast_end; exit;
           #}


          $var ="id=$id\n";

          #var for remote send
          if ($os eq "linux") {
             $t="nc -w 60 $job{'myip'} $job{'port'}";                # sender parameter, options nc, bds-nc, mbuffer
          } else {
             $t="/usr/bin/nc -w 900 $job{'myip'} $job{'port'}";       # sender parameter, options nc, bds-nc, mbuffer
          }
          $var.="src_interface=$t\n";                                # sender par napp-it <= 0.8

          if ($buf > 0) {
            if ($os eq "linux") {
              $t="nc -w 60 $job{'myip'} $job{'port'}";
              $var.="src_interface2=$t\n";                           # sender par napp-it >= 0.9
            } else {
              $t="/var/web-gui/data/tools/nc/nc -b $buf -w 900 $job{'myip'} $job{'port'}";
              $var.="src_interface2=$t\n";                           # sender par napp-it >= 0.9
            }
          }

          if ($job{'send_i'} ne "-I") {
             $job{'send_i'}="-i";                                    # default
          }

          $var.="snap_name1=$newest_src_snap\n";
          $var.="snap_name2=$next_src_snap\n";
          $var.="send_inc=-i\n";
          $var.="snap_rec=$job{'snap_rec'}\n";
          $var.="send_rec=$job{'send_rec'}\n";
          $var.="send_dedup=$job{'send_dedup'}\n";
          $var.="send_i=$job{'send_i'}\n";
          $var.="transfer_zero=$job{'transfer_zero'}\n";

          $var.="send_option=$job{'send_option'}\n";               # ex --raw
          $var.="pv_limit=$job{'pv_limit'}\n";                     # ex -L 300k

          &my_log_id("incremental send","$job{'host'}\: zfs send $job{'send_i'} $job{'send_rec'} $newest_src_snap $next_src_snap | $t");
          &log("$id <br>job-replicate | start sender | init remote send | $var","m");            #monitor

          &my_log_last("start sender\n $var");

          if ($debug=~/1/) { print "\nstart sender $var"; }
          $r = &grouplib_ask_remote_zfs_send("$job{'hostip'}", $var,"","");
          if ($debug=~/1/) { print "\nreturn sender $r\n"; }

          # send error
          if ($r=~/error/) {
              $error="error, request send failed";  $info="error, joblib, request send failed: $r";  &my_fast_end; exit;
          }

          $current{'send_ok'}="1";

          $used=$r;
          if ($r=~/used=/s) {
                $used=~s/.*used=//s;
                &my_log_last("source snap1,2: size=$used");

                # write progress                                        # ex 12M,0
                $t=$used; $t=~s/,0+//;
                $current{'snapsize'}=$t;
                `echo "$t" > /tmp/nappit/$id.progress`;

                # &my_log_id("source snap1,2: size=$used");
          } else {
                $used="";
                &my_log_last("result \n$r");
          }

          # this code will be removed after some more testings
          # opt discard: size=0 !! size always 0 with a delete prior snap on source
          # may be not needed if a failed to read from stream is not regarded as final problem
          #if ($r=~/used=0+,/s) {
          #   if ($job{'transfer_zero'} ne "yes") {
          #      &my_log_last("discard replication $job{'discard0'}, source snap1,2=$used");
          #      &my_log_id("discard replication, source snap1,2=$used");
          #      &my_fast_end;
          #      exit;
          #   }  else {
          #      &my_log_last("replication, source snap1,2=$used ");
          #   }
          #}

          &log("$id <br>job-replicate 741  | | | | $r","m");            #monitor

          &my_monitor;  # monitor receiver
          wait;         # or  ?
          # exit;
     }
}

 ################################
 sub my_local_initial_replication {
 ################################

   my (@t,$t);
   # do not start this job replicate.pl run_123 twice
     $t=`ps axw | grep run_$id | grep -v grep | grep -v color`;  # color_ linux
     @t=();
     @t=split(/\n/,$t);

     # job + fork: 2 allowed
     if (@t > 2) {
         &my_log_id("cancel next, job is running");
         &my_log_last("next replication cancelled<br>job is running");
         exit;
     }

     &my_log_last("start local initial replication");

     my $p;

     # opt delete old snap with this id (restart)
     &my_del_snap($id);

     #create first snap, opt recursiv
     my $snap="$job{'source'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1";
     my $cmd="$ZFS snapshot $job{'snap_rec'} $snap"; $cmd =~s/ +/ /g;

     &my_log_id("cmd","$cmd");
     my $r=`$cmd`;
     if ($r ne "")  {
           &my_log_id("error","$r");
     }
     &log("$id <br>job-replicate 729  | create snap | $cmd | $r ","m");            #monitor


     my $next_dest_snap="$job{'destination'}\@$id\_repli_zfs_$job{'myhost'}\_nr_1";             #backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266

    # job zfs-source size needed in monitor
     $r=`$ZFS list -H -o used $job{'source'}`;
     chomp($r);
     $r=&my_val2mb($r);
     $r="$r\_";
     `echo $r > /tmp/nappit/$id.progress`;

     #debug
   # print "job-replication 741: local init size :$job{'zfs_source_size'}\n$cmd\n$r";

     my $send="$ZFS send $job{'send_option'} $job{'send_dedup'} $job{'send_rec'} $snap"; $send=~s/ +/ /g;
     my $rcv="$ZFS receive $job{'rec_option'} -F $job{'destination'}";                   $rcv =~s/ +/ /g;
     `touch /tmp/$id.receiver`;        # signal for monitor
     $cmd="$send | $rcv 2>&1";

     &my_log_id("cmd","$cmd");

     my $t1=time();
     $r=`$cmd`;
     my $t2=time();

     &log("$id <br>job-replicate 753 | initial send | $cmd | $r ","m");     #monitor
     chomp($r);
     &my_log_id("result","$r");
     &my_log_last("result\n$r");

     &my_postreplication;

     # check for sucessfully created snap
     #read new snaps
     my $t=`$ZFS list -H -r -o name -t snapshot`;   #
     chomp($t);
     my @s=split(/\n/,$t);

     my $ok="0";
     foreach my $s (@s) {
          if ($s eq "$next_dest_snap") {             #ok
              $ok="1";
          }
     }

     if ($ok eq "0") {  # error, no target
              # del source snap
              `$ZFS destroy -f $snap`;
              $error="job-replicate 774: next destination snap $next_dest_snap was not created";
     }

     $t=$t2-$t1;
     my $used=`$ZFS list -H -o used $job{'destination'}`;
     chomp($used);
     $used=&my_val2mb($used);

     my $perf;
     if ($t>0) {
       $perf=$used/$t;
       $perf=sprintf ("%.1f", $perf);
     }

     $info="initial replication finished (time: $t s, size: $used MB, performance: $perf MB/s)"; $info=~s/ +/ /g;

     &my_log_last("$info");
     &my_log_jobs;
     &my_log;
     exit;

 }



 ################################
 sub my_local_delta_replication {
 ################################

   # do not start this job replicate.pl run_123 twice
     my $t=`ps axw | grep run_$id | grep -v grep | grep -v color`;  # color_ linux
     my @t=();
     @t=split(/\n/,$t);

     # job + fork: 2 allowed
     if (@t > 2) {                     # was 1
         &my_log_id("cancel next, job is running");
         &my_log_last("next replication cancelled<br>job is running");
         exit;
     }

     &my_log_last("next incremental replication");

     my $p;

     #step 1 check for newest source dest snaps
     my $newest_src="0";
     my $newest_dest="0";

     my $newest_src_snap="";
     my $newest_dest_snap="";
     my $next_src_snap="";
     my $next_dest_snap="";
     my $last_dest_snap="";

     ($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap) =&my_newest_snap_pair();  # local transfer, no par

     #debug
     #print "\n ## local delta 819 \nns $newest_src_snap\nnd $newest_dest_snap\nxs $next_src_snap\nxd $next_dest_snap\n";
     #exit;

     # get local snaps and opt delete next src snap (kept from unsuccessfull repli)
     my $t=`$ZFS list -H -r -o name -t snapshot -s creation`;  # newest last
     chomp($t);

     my @s=split(/\n/,$t);

     # opt delete next (from failed previous)
     if (($next_src_snap=~/$id\_repli/) && ($id>0))  {
        foreach my $s (@s) {
          if ($s eq "$next_src_snap") {
             `$ZFS destroy -f $s`;
          }
        }
     }

     #create next source snap
     my $cmd="$ZFS snapshot $job{'snap_rec'} $next_src_snap"; $cmd =~s/ +/ /g;
     my $r=`$cmd`;
     &log("$id <br>job-replicate 715 | create snap | $cmd | $r ","m");     #monitor
     &my_log_id("result","$r");
     &my_log_last("result\n$r");

     if ($job{'send_i'} ne "-I") {
          $job{'send_i'}="-i";
     }

     #prepare transfer command

     my $send="$ZFS send $job{'send_option'} $job{'send_dedup'} $job{'send_rec'} $job{'send_i'} $newest_src_snap $next_src_snap";  $send=~s/ +/ /g;
     my $rcv="$ZFS receive $job{'rec_option'} -F $job{'destination'}";                                                   $rcv =~s/ +/ /g;
     my $cmd="$send | $rcv 2>&1 $p";

     &my_log_id("cmd","$cmd");
     &log("$id <br>job-replicate | init local incremental send| $cmd | $r ","m");     #monitor
     #print "job-replication: 853 local delta:\n$cmd\n ";

     my $t1=time();
     `touch /tmp/$id.receiver`;        # signal for monitor
     my $r=`$cmd`;
     my $t2=time();
     # print "job-repli 336 result: $r\n";


     &my_postreplication;


     # ok?
     #read new snaps
     $t=`$ZFS list -H -r -o name -t snapshot -s creation`;       # newest last
     chomp($t);

     @s=split(/\n/,$t);
     @s=reverse(@s);  # newest first to hold last n newest snaps

     $t=$id."_recursiv";
     $t=$current{$t};                                             # yes

     if ($t=~/1|y/i) {
         @s=sort(@s);                 # sort name to allow keep/hold recursive
         if ($debug=~/3/) { print "\nxxxxxxxxx\nsort name\n"; }
     }

     my $ok="0";
     foreach my $s (@s) {
          # print "job-repli 873 snap $s\n";
          if ($s eq "$next_dest_snap") {             #ok
              $ok="1";
          }
     }


     # delete all but newest two pairs
     if ($ok) {
          my $dnr;
          my $dnr2;
          my $dnr3;
          if ($newest_dest_snap=~/_nr_\d+/) {
                   $dnr=$newest_dest_snap;
                   $dnr=~s/.*_nr_//;                                      # current
                   $dnr2=$dnr+1;                                          # next snapnr
                   $dnr3=$dnr-1;                                          # previous

          }
          if (($newest_src_snap ne "")  && ($id>0) && ($dnr>1) && ($newest_dest_snap ne "" ))     {

                     # process local source snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr

                         if ($s=~/^$job{'destination'}/) { next; }        # this is a destination snap

                         #keep not on source
                          if ($s eq $newest_dest_snap)   { next; }
                          `$ZFS destroy -f $s`;
                     }

                     # process local dest snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr
                         if (!($s=~/^$job{'destination'}/)) { next; }     # this is a source snap

                         #daisy rename
                          $r=&my_daisy($s);
                          if ($r) {next; }

                         #keep
                          if (&my_keep($s,$dnr))         { next; }        # process keep on "destination fs"
                          if ($s eq $newest_dest_snap)   { next; }

                          `$ZFS destroy -f $s`;
                     }
         }

         # log
         my $used=`$ZFS list -H -o refer $next_dest_snap`;
         chomp($used);
         $t=$t2-$t1;
         $info="ok, incremental replication finished (time: $t s, new refer: $used)";

         #  print "job repli 77 log ok, log\n";
         #&my_fast_end;

         &my_log_last("$info");
         &my_log_jobs;
          &my_log;
          exit;
     }

     #error:
     $t=$t2-$t1;
     $error="error job-replicate: ";
     $info="next destination snap $next_dest_snap was not created  (runs: $t s)";
     &my_log();
     exit;
 }


####################### get newest source/dest snap-pair
sub my_newest_snap_pair {
#######################     parameter: source snap array on remote transfers

  # print "job repli 927 snappair get newest snappair\n";


  my @snaps=();

  my $t=`$ZFS list -H -r -o name -t snapshot`;         #
  chomp($t);

  my @l=split(/\n/,$t);                                            # local snaps
  push(@snaps,@l);
                                                                   # @snaps: local and remote
  my $newest_src_nr="0";
  my $newest_dest_nr="0";

  my $newest_src_snap="";
  my $next_src_snap="";
  my $next_dest_snap="";
  my $newest_dest_snap="";

  # get newest destination (last successfull replication)
  # print "job-repli 946: search $job{'destination'}\@$id\_repli\n";
  # ex backup3/157_daten@1379949048_repli_zfs_rz-backup3_nr_1

  my $ok="0";
  foreach my $s (@snaps) {                 # destination=local
          chomp($s);
          $s=~s/^ +//;  $s=~s/ .*//;       # snaps with all infos -> name only

          if (!($s=~/^$job{'destination'}\@$id\_repli/)) { next; }  # receiver/my snaps: backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266      (destination - > myhost)
          if (!($s=~/\_nr\_\d+$/)) { next; }                        # repli nr

          #print "job-repli 957: found $s\n";

          $t=$s;
          $t=~s/.*_nr_//;
          if ($t > $newest_dest_nr)  {

               $ok="1";
               $newest_dest_nr=$t;
               $newest_dest_snap=$s;

               $t=$t+1;
               $next_dest_snap=$s;
               $next_dest_snap=~s/\_nr\_.*/_nr_$t/;

          }
  }

  # source with same nr available?                                                                                                 # target snaps   pool/nfs@1375359978_repli_zfs_filer01_nr_2
  ################################

  my $ok2="0";
  if ($ok) {

     my @sn=@_;             # source snaps, empty on local transfers, otherwise remote source snaps
     if (!(@sn > 0)) {      # local transfer source=destination
          @sn=@snaps
     }

     foreach my $s (@sn) {
          $s=~s/ .*//;    # name only
          if ($s=~/^$job{'source'}\@$id\_repli\_.*\_nr\_$newest_dest_nr$/) {                      # source snaps   pool/nfs@137535997_repli_zfs_backup3_nr_2                 (source - > myhost)

                  $s=~s/^ +//;  $s=~s/ .*//;       # snaps with all infos -> name only

                  # previous
                  $pre_newest_src_snap=$newest_src_snap;

                  # newest source must be newest dest always
                  $newest_src_snap=$s;
                  $newest_src_snap==~s/\_nr\_.*/_nr_$newest_dest_nr/;

                  #next source snap = newest dest +1
                  $t=$newest_dest_nr+1;

                  $next_src_snap=$s;
                  $next_src_snap=~s/\_nr\_.*/_nr_$t/;

                  $ok2="1";
                  last;
           }
     }
  } else {
     $ok2="0";
  }


  if ($ok2 eq "0") {       # cancel; error
           &my_log_last("1277, source-dest snap-pair $newest_dest_nr not found: $ok,$ok2");
           $error="error 1275";
           $info ="job-repli, source-dest snap-pair $newest_dest_nr not found, check repli_snaps on source and target. \n";   $info=~s/ +/ /g;

           &my_fast_end;
           exit;
  }


  #return ($newest_src_snap, $next_src_snap)
  $job{'next_src_snap'}=$next_src_snap;
  return($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap);
}

############
sub my_keep {
############
   # keep and hold snaps
   # &my_keep($s,$dnr): snap + newest dest snapnr

   #######################################################
   # keep and hold does not work with rek replication !!!!
   #######################################################

   my $min;
   my $s=$_[0];    # process snaps with newest=last
   my $curr=$_[1]; # current number

   if ($s eq $current{'newest_dest_snap'}) {  return('1'); }                # keep always
   if ($s eq $current{'last_dest_snap'})   {  return('1'); }

   #get filesystem
   $t=$s;
   $t=~s/\@.*//s;

   if ($t ne $keepfs) {                   # first or new fs, to allow recursive replications
        $keepfs=$t;
        @keeph=();
        @keepd=();
        @keepm=();
        @keepy=();
        $holdnr="0";
   }


   my $t=$id."_keep";                     # keep is a commalist like hours:24,days:32,months:12,years:2,.00 (last is char)
   $t=$current{$t};
   $t=~s/ +//g;

   my $hold=$id."_hold";                  # keep is a commalist like hours:24,days:32,months:12,years:2,.00 (last is char)
   $hold=$current{$hold};

   if ($debug=~/3/) { print "\n1469 sub my_keep process keep/hold snap:\n  $s\n"; print "  keep $t, hold $hold\n"; }

   my @keep=split(/,/,$t);
   my (@char,$char,$hours,$days,$months,$years,$n,@w,$r,$snapdate,$lastnr);


   $snapdate=`$ZFS list -H -o creation $s`;  # ex Sun Sep 15  0:05 2013
   chomp($snapdate);

   foreach my $t (@keep) {
      if ($t eq "") { next; }
      if ($t=~/hour.*\d/) {
        $t=~s/\D//g;
        if ($t == 0) { $t="1"; }               # hour 1..24
        $hours=$t;
        next;
      }
      if ($t=~/day.*\d/)   { $t=~s/\D//g; $days=$t;   next; }
      if ($t=~/month.*\d/) { $t=~s/\D//g; $months=$t; next; }
      if ($t=~/year.*\d/)  { $t=~s/\D//g; $years=$t;  next; }


      #char
      push (@char,$t);
   }

   # 1. check keep char with rename (keep outside retention)
   ##################
   if (@char>0) {
      foreach my $p (@char) {
        if ($p eq "")  { next; }
        if ($s=~/$p$/) {
             $n=$s;
             $n=~s/.*\_//;
             &my_log_id("1499 keep nr $n","keep $t: match rule: $p");
             #rename to keep it even when rule changes

             $t=$s; $t=~s/repli/keep/;
             $t=~s/\_nr/\_was/;                # or autodelete will happen on _nr_n
             `$ZFS rename $s $t`;
              if ($debug=~/3/) { print "1505 keep char >>>>> 1505 rename $s\n$t\n\n"; }
             return('1');                      # skip delete
        }
      }
  }

  @w=();
  if (($hours>0) || ($days>0) || ($months>0) || ($years>0)) {
     $r=$snapdate;
     chomp($r);
     $r=~s/ +/ /g;

     @w=(); @w=split(/ +/,$r);

     #month in w[1], day in w[2], year in w[4]
     # month can be mai Mai May (depends on localisation)
     $w[1]=~s/Jan/01/i;     $w[1]=~s/Feb/02/i; $w[1]=~s/Mar/03/i; $w[1]=~s/Apr/04/i;
     $w[1]=~s/Ma[ijy]/05/i; $w[1]=~s/Jun/06/i; $w[1]=~s/Jul/07/i; $w[1]=~s/Aug/08/i;
     $w[1]=~s/Sep/09/i;     $w[1]=~s/^O.*/10/i; $w[1]=~s/Nov/11/i; $w[1]=~s/De.*/12/i;

     if ($w[2] < 10) {         # 3 -> 03
       $w[2]="0$w[2]";
       $w[2]=~s/^0+/0/;
     }

     $w[3]=~s/:.*//;    # hour 0..23 only
     if ($debug=~/3/) { print "  1527 snapdate $snapdate from zfs get creation: $r\n    day dd.mm.yyyy hour = $w[0] $w[2].$w[1].$w[4] hour: $w[3] vs today $sys{'mday'}.$sys{'mon'}.$sys{'year'} hour $sys{'hour'}\n"; }
  }


  # 2. keep on snap per last 1..n years
  if ($years > 0) {
      # how old is snap in years
      $t=0;    #
      for (my $y=$sys{'year'};$y>=$w[4];$y--) {           # this=1
              $t++;
      }

      if ($debug=~/3|y/) { print "  1539 process years=$years with snap-age in years (this=1): $t\n"; }

      # keep year
      if (($keepy[$t] ne "1") && ($t<=$years)) {            # global array of years, this year keepy[1]
             $keepy[$t]="1";              # remember

             $t=$s;
             $t=~s/_zfs\-..\d+_/_zfs_/;   # opt remove old rules, old script had used hn30s for 30 snaps!
             $t=~s/_zfs_/_zfs-ky$w[4]\_/;
             if ($s ne $t) {
                 `$ZFS rename $s $t`;
                 if ($debug=~/3|y/) { print "      1550 rename and keep year  $w[4] $s\n        to $t\n"; }        # keep the oldest from this year
             }  else {
                 if ($debug=~/3|y/) { print "      1553 keep year $w[4] $s\n"; }                         # keep the oldest from this year
             }
             return('1');       # skip delete
      }  else {
             if ($debug=~/3|y/) { print "    1558 year $w[4] is already kept or > maxmyears\n"; }
      }
  }



  # 3. keep on snap per last 1..n months     (debug 3c)
  if ($months > 0) {
     #how old is snap in months
#$w[4]--; # debug

     if ($debug=~/[3789m]/) { print "\n1756 process $s\n months=$months on $w[1] $w[4] \n"; }


     $t=0;    #
     for (my $y=$w[4];$y<=$sys{'year'};$y++) {                   # snapyear to this year

        if (($y==$w[4]) && ($y != $sys{'year'})) {               # old year of snap
          for (my $m=$w[1];$m<12;$m++) {                         # til end of year
            $t++;
          }
          if ($debug=~/7/) { print " 1762 snapyear to end year $w[4]: $t\n"; }
          next;
        }

        if (($y>$w[4]) && ($y<$sys{'year'})) {                   # years between
          $t=$t+12;
          if ($debug=~/8/) { print " 1771 years between: $t\n"; }
          next;
        }

        # snap is from older years
        if ($w[4] != $sys{'year'}) {
          for (my $m=1;$m<=$sys{'mon'};$m++) {                   # til this month and old year
            $t++;
          }
          if ($debug=~/9/) { print " 1777 this year to this month $sys{'mon'}: $t\n"; }

        }  else {

          # snap is from this year

          for (my $m=$w[1];$m<$sys{'mon'};$m++) {               # til this month and snap this year
            $t++;
          }
          if ($debug=~/9/) { print " 1788 this year to this month $sys{'mon'}: $t\n"; }
        }
      }

      if ($debug=~/[3789]|m/) { print " 1783: month-diff=$t\n\n"; }
      $t++; # diff=0 -> $t=1

      # keep month
      if (($keepm[$t] eq "") && ($t<=$months)) {
             $keepm[$t]="1";    # remember

             $t=$s;
             $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
             my $yr=substr($w[4],2,2);
             $t=~s/_zfs_/_zfs-km$yr$w[1]\_/;
             if ($s ne $t) {
               `$ZFS rename $s $t`;
               if ($debug=~/3|m/)  { print "    1583 rename and keep month $w[1] $w[4] $s\n$t\n\n"; }
             } else {
               if ($debug=~/3|m/)  { print "    1585 keep month $w[1] $w[4] $s\n"; }                         # keep the oldest from this year
             }

             return('1');       # skip delete
      } else {
             if ($debug=~/3|m/) { print "    1590 month $w[1] $w[4] (month age=$t) is already kept or > max months\n"; }
      }
  }

  # ex Sun Sep 15  0:05 2013
  # 4. keep one snap per last 1..n days
  if ($debug=~/3|d/) { print "  1688 process $s with $snapdate and days $days with date=$w[2].$w[1].$w[4]\n"; }
  if ($days > 0) {
      $t=&date2diff("$w[2].$w[1].$w[4]","$sys{'mday'}.$sys{'mon'}.$sys{'year'}");
      if ($debug=~/3|d/) { print "    1553 process days $days\: $w[2].$w[1].$w[4] >> $sys{'mday'}.$sys{'mon'}.$sys{'year'}\n 1553 diff=$t\n"; }

      if ($t <= $days)   {
         # keep day

         $r="$w[1]$w[2]";
         if ($keepd[$r] eq "") {
             $keepd[$r]="1";            # remember
             if ($debug=~/3|d/) { print "    1524: process day: $w[2] with diff=$t\n"; }

             $t=$s;
             $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
             $t=~s/_zfs_/_zfs-kd$w[1]$w[2]\_/;
             if ($s ne $t) {
                `$ZFS rename $s $t`;
                if ($debug=~/3|d/)  { print "      1615 rename and keeo day $w[2] $s\n$t\n"; }
             }  else {
                 if ($debug=~/3|d/) { print "      1617 keep day $w[2] $s\n"; }                         # keep the oldest from this year
             }

             return('1');        # skip delete
         }
      }  else {
#debug
#print "!\n"; exit;
             if ($debug=~/3|d/) { print "    1558 day $w[2] ($t) is already kept\n"; }
      }
  }

  # 5. keep last n snaps per hour
  if ($hours > 0) {

      my $h1=&date2dop("$sys{'mday'}.$sys{'mon'}.$sys{'year'}") * 24;                # passed number of hours now
      my $h2=&date2dop("$w[2].$w[1].$w[4]") * 24;                                    # passed number of snap hours

      $h1=$h1+$sys{'hour'};    # hour value of now
      $h2=$h2+$w[3];           # hour value of snap

      $r=$h1-$h2;              # hour diff

      if ($debug=~/3|s/) { print "    1619 process hours $hours with diff=$r\n"; }
      if ($r <= $hours) {

              # @keeph: pos==data
              if ($keeph[$r] eq "") {
                   $keeph[$r]="1";    # remember
                   if ($debug=~/3|s/) { print "    1626: process hour: $r\n"; }

                   $t=$s;
                   $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
                   $t=~s/_zfs_/_zfs-kh$r\_/;
                   if ($s ne $t) {
                     `$ZFS rename $s $t`;
                     if ($debug=~/3|s/) { print "    1643 rename and keep hour $r $s\n  to $t\n"; }
                   }  else {
                     if ($debug=~/3|s/) { print "    1553 keep hour $r  $s\n"; }                         # keep the oldest from this year
                   }

                   return('1');        # skip delete
              }
      }  else {
             if ($debug=~/3|s/) { print "    1654 diff hour $r  > $hours \n"; }
      }

      #debug
      #return(1);
  }

  #######################################
  # 6. check hold (days ex 8 or snaps ex 8s)
  #######################################
  $t=$id."_hold";
  $t=$current{$t};
  my $hld=$t;
  $hld=~s/\D//g;

  if ($debug=~/3|h/) { print "  1664 process hold $t\n"; }

  # number of snaps ex 8s to hold additionally
  if ($t=~/s/i) {                     # number of snaps
                                      # use _nr as reference
       $n=$s;
       $n=~s/.*nr\_//;                # current snap nr

       if (($hld > 0) && ($n>0) && ($holdnr<=$hld))  {
          $lastnr=$current{'last_dest_snap'};
          $lastnr=~s/.*nr\_//;
          $lastnr=$lastnr-1;                                # =newest

          if ($n >= ($lastnr-$hld) ) {       # more snaps than max to hold
                  if ($debug=~/3f/) { print "1673 process hold number: $n -> $hld (last: $lastnr)\n"; }

                  $holdnr++;
                  $t=$s;
                  $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
                  $t=~s/_zfs_/_zfs-hn$holdnr\_/;
                  if ($s ne $t) {
                     `$ZFS rename $s $t`;
                     if ($debug=~/3|h/) { print "      1678 hold $holdnr snap and rename $s\n      to $t\n"; }
                  } else {
                     if ($debug=~/3|h/) { print "      1680 hold $holdnr snap $s\n"; }                         # keep the oldest from this year
                  }
                  return('1');        # skip delete
            }  else {
                  if ($debug=~/3|h/) { print "    1654 max snaps $hld already hold\n"; }
            }

       }

  } else {

       if ($hld > 0) {

           $r=$snapdate;
           chomp($r);
           $r=~s/ +/ /g;

           @w=();
           @w=split(/ /,$r);

           #month in w[1], day in w[2], year in w[4]
           # month can be mai Mai May (depends on localisation)
           $w[1]=~s/Jan/01/i;     $w[1]=~s/Feb/02/i; $w[1]=~s/Mar/03/i; $w[1]=~s/Apr/04/i;
           $w[1]=~s/Ma[ijy]/05/i; $w[1]=~s/Jun/06/i; $w[1]=~s/Jul/07/i; $w[1]=~s/Aug/08/i;
           $w[1]=~s/Sep/09/i;     $w[1]=~s/O.*/10/i; $w[1]=~s/Nov/11/i; $w[1]=~s/De.*/12/i;

           $r=&date2diff("$w[2].$w[1].$w[4]","$sys{'mday'}.$sys{'mon'}.$sys{'year'}");

           #debug
           # &my_log_id("snapdiff","$s age: $days, hold $t days");

           if ($r < $t) {            # age < hold

               # log
               if ($debug=~/3|h/) { print "    1718: process days: $r\n"; }

               $t=$s;
               $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
               $t=~s/_zfs_/_zfs-hd$hld\_/;
               if ($s ne $t) {
                     `$ZFS rename $s $t`;
                     if ($debug=~/3|h/) { print "      1719 hold days $hld, rename $s\n      to $t\n"; }
               } else {
                     if ($debug=~/3|h/) { print "      1721 hold days &hld $s\n"; }                         # keep the oldest from this year
               }

               return('1');                      # skip delete
           }
       }
  }



  # ok, no extra hold/keep
  if ($debug=~/3|h/) { print "xxxx 1727 destroy snap $s\n"; }
  return('0');
}

 #################  # on initial replication
 sub my_del_snap {
 ################# par jobid, nr
 # receiver/my snaps: backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266
 # sender snaps  : www/nfs@1330674968_repli_zfs_backup3_nr_266

   my ($n);
   my $id=$_[0];
   if (!($id=~/^\d+$/))  { print "job-repli delsnap 184: no id $id\n"; return; }

   my $nr=$_[1];     # when nr, delete all but two newest
   if ($nr > 0) {
       $nr=$nr-2;
   }



   my $t=`$ZFS list -H -r -o name -t snapshot`;
   my @s=split(/\n/,$t);
   foreach my $s (@s) {

        if (!($s=~/\@$id\_repli/)) { next; }     # not my id
        if (!($s=~/_repli/))       { next; }     # only repli snaps

        # without nr
        if ($nr eq "") {
              `$ZFS destroy \'$s\'`;
        } else {
               if (!($s=~/_nr_\d+$/))       { next; }     # only nr repli snaps
               $n=$s;
               $n=~s/.*nr\_//;
               if (!($n>0)) { next; }

               if ($n<=$nr) {                             # keep last two
                   `$ZFS destroy \'$s\'`;
               }
        }
   }
}


########## end quickly
sub my_fast_end {
##########

   #&my_log_id("my fast_end: cancel all");

   if (-f "/tmp/nappit/$id.running")    {  `sudo rm /tmp/nappit/$id.running"`;   }
   if (-f "/tmp/nappit/$id.receiver")   {  `sudo rm /tmp/nappit/$id.receiver"`;  }

   &my_end_all;        # kill all at once
   exit;
}

############ # cancel job and fork
sub my_end_jobs {
############

  my $r=`ps axw | grep _$id | grep -v grep | grep -v color`;  # color_ linux
  my ($k,$p,@p,$t,$l);

  @p=split(/\n/,$r);
  foreach my $p (@p) {
      $t.="kill job $p; ";
      $p=~s/^ +//;
      $p=~s/ .*//;
      $k.="kill -9 $p >/dev/null 2>&1; ";
  }

  # &my_log_last("exe: $t");
  if ($_[0] eq "l") { return($k); }  # end_all
  $r=`$k`;
  exit;          # must be last action
}

########### kill proc and jobs as a single task ex to cancel on snaps=0
sub my_end_all {
###########
      my $l=&my_end_jobs('l');
      $l.=&my_end_proc('l');
      #&my_log_last("Kill: $l");
      &my_log;
      `$l`;
      exit;
}

############ end all nc and zfs proc
sub my_end_proc {
###########

  # parameter l=return list of kill
  # kill nc and opt zfs receive

  # debug
  # return;

   my $r=`ps axw`;
   my @p=split(/\n/,$r);
   my ($ok,$t,$l);

   # step 1, kill nc
   foreach my $p (@p) {
      if ($p=~/grep/)              { next; }
      if ($p=~/httpd|server/)      { next; }
      if (!($p=~/\/nc|zfs|job\-replicate/))      { next; }

      $ok="0";

      #send (locally)
      if ($id ne "") {
         if ($p=~/zfs send .*$id/) { $ok="1"; }
      }

      #nc port
      if ($job{'port'} ne "") {
        if ($p=~/nc .* \-p $job{'port'}/) {  $ok="1"; }
      }

      # job
      # is killed in my_log as last action

      if ($ok ne "1") { next; }

      #get ps
      $t=$p;
      $t=~s/^ +//;
      $t=~s/ .*//;

      &my_log_last("kill proc $p ($t)");
      if ($_[0] eq "l") { $l.="kill -9 $t >/dev/null 2>&1; "; next; }  # end_all
      $r=`kill -9 $t >/dev/null 2>&1`;

   }

   sleep(1);
   # step 2 kill zfs  receive
   foreach my $p (@p) {
      if ($p=~/grep/)                            { next; }
      if ($p=~/httpd|server/)                    { next; }
      if (!($p=~/\/nc|zfs|job\-replicate/))      { next; }

      $ok="0";

      #receive: nc -d -l -p $job{'port'} | $ZFS receive -F $job{'destination'}
       if ($job{'destination'} ne "") {
          if ($p=~/zfs receive .*$job{'destination'}$/) { $ok="1"; }
       }
      if ($ok ne "1") { next; }

      #get ps
      $t=$p;
      $t=~s/^ +//;
      $t=~s/ .*//;

      &my_log_last("kill proc $p ($t)");
      if ($_[0] eq "l") { $l.="kill -9 $t >/dev/null 2>&1; "; next; }  # end_all

      $r=`kill -9 $t`;

   }

   if ($_[0] eq "l") { return($l); }         # list of kill

}

## cancel job
sub my_cancel  {
###############

   $info="job $id.cancel requested in job-replicate 696";

   # kill all processes with jobid
   # kill receive processes (terminates remote send ?)

    &my_log_last("my cancel, kill all processes");
    if (-f "/tmp/nappit/$id.running")    {  `sudo rm /tmp/nappit/$id.running"`;  }
 }

 ## delete job
 sub my_delete {
 ##############
 my ($r,$t);
      if ($id ne "") {

        &my_cancel;

        $t="rm /var/web-gui/_log/jobs/$id.*";
        $r=`$t`;

        $t="rm /tmp/nappit/$id.*";
        $r=`$t`;

        print "1297 #### job-replication delete\n$t\n$r\n";

      } else {
        print "job-replication delete, no id !! ########### \n";
      }

      if (-f "/tmp/nappit/$id.delete")    {  unlink("/tmp/nappit/$id.delete");   }

      $info="delete job $id in job-replicate";               # no log.id!
      &my_fast_end;
}

########### log detailed info for last job (klick on Date of last exec)
sub my_log_last {
##########
   &task_gettime;
   my $t=$_[0];

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);

                  my $l="$p0[3] line $p0[2]";             $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $l.="<br> <- $p1[3] line $p1[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $l.="<br> <- $p2[3] line $p2[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $l.="<br> <- $p3[3] line $p3[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $l.="<br> <- $p4[3] line $p4[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   $l=~s/main::/\&/gs;
   $l=~s/br> +/br> /g;
   $l=~s/\s+/ /g;

   $l=~s/\n/<br>/g;
   $t=~s/\n/<br>/gs;


   my $log="/var/web-gui/_log/jobs/$id.last";
   open (PF, ">>$log");
     flock(PF, LOCK_EX);
     print PF "$current{'timestamp'}\t$l\t$t\n";
   close (PF);
}


########### write log
sub my_log_jobs {
###############
 my (@v,$v,$n,$l,$f1,$f2,@data,$t);
 my @p1=caller(0);

 $info=~s/\n.*//s;                                        # first line

  $l=$error;
  if ($l eq "") { $l="ok"; } else { $l="error"; $info="$info $error"; }

  if ($error=~/smb add share failed/) {
       $l="info";
  }
  $l=~s/\n/ /gs;

  #stage 1: log general jobs.log
  ###########################

    my $log="/var/web-gui/_log/jobs/jobs.log";
    my $lock="/var/web-gui/_log/jobs/jobs.lock";
    open (X,"<$lock");
          flock(X, LOCK_EX);
          #
          open (F,"<$log");
                flock(F, LOCK_EX);
                @data=<F>;
          close (F);

          #unshift   $l=ok/error
          unshift(@data,"$l; replication $id; time: $current{'timestamp'} line: $p1[2];info $info\n");

          if (@data > 50) {
                     $t=@data-50;
                     splice(@data,50,$t);
          }

         $t=join("",@data);
         open (PF, ">$log");
            flock(PF, LOCK_EX);
            print PF "$t";
         close (PF);
   close (X);
}

###########   ## log last run in jobfile /_log/jobs/id...job
sub my_log {  # always last action!
###########
   my (@v,$v,$n,$l,$f1,$f2,@data);
   my @p1=caller(0);

  # print "job-repli 1237 - call log id=$id action=$action\n";

  # line 1: lastrun = ok or err
  # line 2: time of last run
  # line 3+: info

  $l=$error;
  if ($l eq "") {
       $l="ok";
  } else {
       $l="error";
       $info="$info $error";
  }

  if ($error=~/smb add share failed/) {
       $l="info";
  }

  $l=~s/\n/ /gs;

  #rm progress
  if (-f "/tmp/nappit/$id.progress") {  unlink("/tmp/nappit/$id.progress"); }

  $info=~s/\n.*//s;                                        # first line


  if ($info=~/not found/) { $info.=". <br> Check filesystems and group membership, opt delete newest target snap if second new is available."; }

  my $r=`ps axw | grep web-gui`;

 # do not update logfile on delete
 if ($action ne "delete") {
       # stage 3 id.log newest at top
       ###################
       &my_log_id("$l","my_log end 1551: $info");
       &my_log_last("$l","my_log end: $info\n$r");

       #monitor
       &log("$id $l <br>job-replicate 1331 | end log | $info | ","mr");     #monitor
  }

  if (-f "/tmp/$id.receiver")          { `rm /tmp/$id.receiver`; }                    # logfile receiver started

}

##############  log to id.log
sub my_log_id {
#############
my $err=$_[0];
my $data=$_[1];
$data=~s/ +/ /g;
my @data;
&task_gettime;

my @p1=caller(0); # this:
my @p2=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask
my @p3=caller(2);

my $log="/var/web-gui/_log/jobs/$id.log";
  open (F,"<$log");
        flock(F, LOCK_EX);
        @data=<F>;
  close (F);

  #unshift   $l=ok/error
  unshift(@data,"$err; time: $current{'timestamp'} line: $p1[2]; $data\n");

  if (@data > 300) {
             $t=@data-300;
             splice(@data,300,$t);                 # delete from 50+
  }

 $t=join("",@data);
 open (PF, ">$log");
    flock(PF, LOCK_EX);
    print PF "$t";
 close (PF);

}

########################
sub my_postreplication {
########################
  my ($r,$type,$s,$abe,@data,$t,@t);

  # rec: set mountpoint to job{'destination'}
  if ($job{'send_rec'}=~/R/)  {
          &my_log_last("rec=$job{'send_rec'},$job{'recursiv'}<br>set mountpoint of $job{'destination'} to /$job{'destination'}");
          `$ZFS set mountpoint=/$job{'destination'} $job{'destination'}`;
  }

  if ($os eq "linux") {  print "Linux, no post actions\n"; return; }        # no post

  &my_log_last("begin my_postreplication");

  $r = &grouplib_ask_remote_snaplist("$job{'hostip'}");   #commalist
  @t=split(/,/,$r);

  # debug
  #$r=~s/,/<br>/gs;
  #&my_log_last("snaps on $job{'host'}, next: $job{'next_src_snap'}<br>$r");

  foreach my $t (@t) {
          if ($job{'next_src_snap'} eq "")       { &my_log_last("next_src_snap=n.a.");  last; }
          if (!($t=~/$job{'next_src_snap'}/))    { next; }
          &my_log_last("next_src_snap from host $job{'host'}<br>$t");
  }

  # update last in $id.par
  &task_gettime;
  my $log="/var/web-gui/_log/jobs/$id.par";
  open (F,"<$log");
        flock(F, LOCK_EX);
        @data=<F>;
  close (F);
  foreach my $t (@data) {
         if ($t=~/^last=/)   { $t="last=$current{'logtime'}\n"; }
  }
  $t=join("",@data);
  open (PF, ">$log");
    flock(PF, LOCK_EX);
    print PF "$t";
  close (PF);
  &my_log_last("1522 update id.par: last=$current{'logtime'}");
  ###############################


  # step 1: load destination ZFS and ZFS below
     my $zfs=`$ZFS list -H -o name`;   # all ZFS names
     chomp($zfs);
     my @t=();
     @t=split(/\n/,$zfs);

     #collect ZFS
     my @zfs=();
     foreach my $p (@t) {
            chomp($p);
            if ($p ne $job{'destination'}) { next; }

            $type=`$ZFS list -H -o type`;                     # do not post-process volumes
            chomp($type);
            if ($type eq "volume") { next; }

            push (@zfs,$p);
            #print "found ZFS: $p\n";
     }

  # step 2: set all shares to readonly=off and perm to 777   (if reshared)
     foreach my $zfs (@zfs) {

             $r=`$ZFS set readonly=off \'$zfs\' 2>&1`;
             &my_log_last("zfs set readonly=off $zfs\n$r");

             if ($job{'dest_sharename'} ne "") {         # must be set or may get access denied
                 if ($job{'dest_shareguest'} =~/enable|yes/) {
                      $r= `chmod 777 \'/$zfs\'`;
                      #print "set perm to 777 \'$zfs\'\n$r";
                 } else {
                      $r= `chmod 755 \'/$zfs\'`;
                      #print "set perm to 755 \'$zfs\'\n$r";
                 }
             }
     }

   # step 3 - set all shares to off
     foreach my $zfs (@zfs) {
             $r=  `$ZFS set sharesmb=off \'$zfs\'`;
             $r=  `$ZFS inherit -r sharesmb \'$zfs\'`;      # rec

             $r.= `$ZFS set sharenfs=off \'$zfs\'`;
             $r=  `$ZFS inherit -r sharenfs \'$zfs\'`;      # rec
             &my_log_last("zfs set nfs/smb shares=off $zfs\n$r");
     }
  # step 4 reshare it optionally
     my $cmd;
     if ($job{'dest_sharename'} ne "") {
                  foreach my $zfs (@zfs) {

                         $t="$zfs";
                         $t=~s/\//_/g;  # a/b -> a_b           # sharename=dataset name

                         #overwrite dest_sharename
                                    $s=$job{'destination'};
                                    $s=~s/\//_/g;  # a/b -> a_b

                                    $t=~s/$s/$job{'dest_sharename'}/;

                         #abe?
                         if ($job{'dest_shareabe'} =~/enable|yes/) {
                           $abe=",abe=true";
                         } else {
                           $abe="";
                         }

                         ####
                         if ($sys{'auto_os'} eq "Solaris 11") {

                              if ($job{'dest_shareguest'} =~/enable|yes/) {
                                $r=&exe("$ZFS set share=name=$t,prot=smb$abe,guestok=true $in{'zfs'}");
                              } else {
                                $r=&exe("$ZFS set share=name=$t,prot=smb$abe $in{'zfs'}");

                              }
                              $r=&exe("zfs get -H -o value sharesmb $in{'zfs'}");
                              if ($r eq "off") {           # additional: sharesmb=on
                                  $r=&exe("$sys{'sbinpath'}zfs set sharesmb=on $in{'zfs'}");
                              }

                        } else {

                              ####

                              #$cmd="$ZFS set sharesmb=name=$vars{'dest_sharename'}$t";
                              if ($job{'dest_shareguest'} =~/enable|yes/) {
                                   $r=`$ZFS set sharesmb=name=$t,guestok=true$abe \'$zfs\'`;

                                  # $cmd.=",guestok=true \'$zfs\'; ";
                              } else {
                                   $r=`$ZFS set sharesmb=name=$t$abe \'$zfs\'`;
                                  # $cmd.=" \'$zfs\'; ";
                              }
                         }
                         &my_log_last("zfs set nfs/smb shares=on $zfs\n$r");
                  }
     }

     # step 5 set to read only if writable is not yes
     #print "\n";
     foreach my $zfs (@zfs) {
             if ($zfs=~/\/ROOT\//) { next; }                       # BE, do not set readonly
             if ($job{'writable'} =~/enable|yes/) {

               $r=`$ZFS set readonly=off \'$zfs\' 2>&1`;
               &my_log_last("my postreplication, zfs set readonly=off $zfs\n$r");

             } else {

               $r=`$ZFS set readonly=on \'$zfs\' 2>&1`;
               &my_log_last("my postreplication, zfs set readonly=on $zfs\n$r");
             }
     }

     # opt post job
     if (-f "/var/web-gui/_log/jobs/$id.post") {
       &my_log_last("execute post shellscript $id.post");
       $r=`sudo sh /var/web-gui/_log/jobs/$id.post`;
       print "$r\n";
       &my_log_last("result of post shellscript\: $r");
     }

      # remove ESXi snaps
      if ($job{'esxi_ip'} =~/\d/) {
        $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl remove_$id`;   # remove ESXi snaps, can happen to fail on some snaps due "too busy"
      }

     &my_log_last("end my postreplication");

}

######################
sub my_wait_and_check {  # monitor wait and read cancel/del
#######################
    my $t=$_[0];
    while ($t>0) {
       if (-f "/tmp/nappit/$id.cancel")    { &my_cancel; &my_fast_end;  exit; }
       if (-f "/tmp/nappit/$id.delete")    { &my_delete; exit; }
       sleep (2);
       $t=$t-2;
    }
}



###########
sub my_par {
###########     # read current par_* to %job

    #my $id=$_[0];
    $current{'port'}                 = 81;

    #&task_read_jobpar($id);            # need only current id

    $t=$id."_host";
    $job{'host'}              = $current{$t};                 # senders hostname               ex box3

    $t=`hostname`; chomp($t);
    $job{'myhost'}            = $t;                           # receiver (my hostname )        box1

    $t=$id."_source";
    $job{'source'}            = $current{$t};                 # source ZFS                     tank/zfs1

    $t=$id."_destination";
    $job{'destination'}       = $current{$t};                 # dest ZFS                       tank2/zfs1

    $t=$id."_writable";
    $job{'writable'}          = $current{$t};                 # dest ZFS writable             tank2/zfs1

    #recursiv
    $t=$id."_recursiv";
    if ($current{$t} eq "yes") {
      $t=$id."_snap_rec";
      $job{'snap_rec'}          = "-r";                       # ** snap recursiv               -r or ""

      $t=$id."_send_rec";
      $job{'send_rec'}          = "-R";                       # ** send recursiv               -R or ""
    }

    $t=$id."_src_interface";
    $job{'src_interface'}     = $current{$t};                 # source (sender)                sudo /usr/bin/nc -w 10 172.16.11.28 58309

    $t=$id."_dest_interface";
    $job{'dest_interface'}    = $current{$t};                 # destination (receiver)         sudo /usr/bin/nc -d -l -p 58309                 ## no wait !!

    #obsolete
    $t=$id."_local_running_ps";
    $job{'local_running_ps'}  = $current{$t};                 # on local/receiver               nc .* -p 58309++zfs receive .*tank2/zfs1

    $t=$id."_dest_sharename";
    $job{'dest_sharename'}    = $current{$t};                 #                                 'name' or ""

    $t=$id."_dest_shareguest";
    $job{'dest_shareguest'}   = $current{$t};                 # guestok?                        enable or ""

    $t=$id."_dest_shareabe";
    $job{'dest_shareabe'}   = $current{$t};                   # abe?                            enable=yes

    $t=$id."_send_dedup";
    $job{'send_dedup'}        = $current{$t};                 # send dedup                     -D or ""

    $t=$id."_send_option";
    if (!($current{$t}=~/[a-zA-Z0-9]/)) {$current{$t}="";  }
    $job{'send_option'}        = $current{$t};                # send option like raw

    $t=$id."_rec_option";
    if (!($current{$t}=~/[a-zA-Z0-9]/)) {$current{$t}="";  }
    $job{'rec_option'}        = $current{$t};                # receive option

    $t=$id."_pv_limit";
    if (!($current{$t}=~/[a-zA-Z0-9]/)) { $current{$t}="";  }
    $job{'pv_limit'}        = $current{$t};                  # throughput limit ex -L 300k


    $t=$id."_port";
    $job{'port'}   = $current{$t};                            # port

    $t=$id."_text1";
    $job{'text1'}   = $current{$t};                            # port


    $t=$id."_hostip";
    $job{'hostip'}        = $current{$t};                     # sender ip

    $t=$id."_esxi_ip";
    $job{'esxi_ip'}        = $current{$t};                     # esxi_ip

    $t=$id."_send_i";
    $job{'send_i'}        = $current{$t};                     # zfs send -I or -i

    #optional job port and hostip (from old jobs)
    if ($job{'port'} eq "") {
        $job{'port'}=$job{'text1'};
        $job{'port'}=~s/.*\-//;                               # 192.168.88.74-58429
    }

    if ($job{'hostip'} eq "") {
        $job{'hostip'}=$job{'text1'};
        $job{'hostip'}=~s/\-.*//;                             # 192.168.88.74-58429
    }
    if (!($job{'hostip'}=~/^\d+\.\d+\.\d+\.\d+$/)) {$job{'hostip'}=$job{'host'}; }

    $t=$id."_transfer_zero";                                  # transfer_zero
    $job{'transfer_zero'}   = $current{$t};

    #daisy
    $t=$id."_daisy";                                           # daisy chain repli to jobid
    $job{'daisy'}   = $current{$t};
    $job{'daisy'}=~s/\D//g;
    if (length($job{'daisy'})  ne "10") { $job{'daisy'}=""; }

    #debug
      #print "\n #######  job-repli pars 702 ####### id=$id\n";
      print "2489: $job{'daisy'}\n";

}

###### debug
sub print_my_par {
#################
     my ($k,$v,$t);
     my $id=$_[0];

     while (($k,$v) = each %current) {
            if (!($k=~/^$id\_/))    { next; }
            $t.= "current $k\=$v\n";
     }



     #print par
     while (($k,$v) = each %job) {
            $t.= "job $k\=$v\n";
     }
     return ($t);
}

 ###############
 sub my_val2mb  {  #hide:
 ###############

      my $w1=$_[0];
      if ($w1=~/K/) { $w1=~s/K//; $w1=$w1/1000;    }
      if ($w1=~/M/) { $w1=~s/M//; $w1=$w1*1;       }
      if ($w1=~/G/) { $w1=~s/G//; $w1=$w1*1000;    }
      if ($w1=~/T/) { $w1=~s/T//; $w1=$w1*1000000; }
      return ($w1);
 }

 #end
 1;
