#!/usr/bin/perl
# (c) napp-it 2012-2023
# 27.may.2020, delzero: recheck used for each snap
# 13. apr 2023 esxi snaps

# debug: run from cli    perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/job-snap.pl run_id_time
# last: path problem
# fix with delzero

use strict;
use Fcntl qw(:flock);                                          # filelocking
$|=1;                                                          # keine Pufferung, no buffering

use vars qw($par $id $action $f $t @t $p @p %sys %cfg %current $error $info $message $text $subject $smtp $ref);

$par=$ARGV[0];                                                 # action_$job time, for debugging purpose, you can call manually via cli with par: run_id currenttimestamp
                                                               # ex:  perl job-snap.pl run_1357655102 135796743
 # common lib
 use lib "/var/web-gui/data/napp-it/CGI";                      # add to perl search path for libs ex. Expect.pm and others
 require "/var/web-gui/data/wwwroot/cgi-bin/admin-lib.pl";

 $f=`uname -a`;
 if ($f=~/Linux/){
       $f="linux";

 } else {
       $f="illumos";
       $sys{'sbinpath'}='/usr/sbin/';
 }

 require "/var/web-gui/data/napp-it/zfsos/_lib/scripts/task-lib.pl";        #
 #require "/var/web-gui/data/napp-it/zfsos/_lib/$f/zfslib.pl";

 if (!(defined($par))) {
     &log("job-snap:  missing parameter");
     exit;
 }

 ############

 #par: action_job timestamp
 $par=~s/ .*//;

 # parameter: action_$job time
 if (!($par =~/\w+\_\d+$/)) {
     print "job-snap30: wrong par $par";
     exit;
 }
 @t=();
 @t=split(/\_/,$par);    # action job
 $action=$t[0];
 $action=~s/[^a-z]//g;

 $id=$t[1];
 &task_gettime;

  &my_logid("start; job-snap.pl $ARGV[0]");
 #############################################

 # get par -> %current
 &task_read_jobpar($id);

# run?
if ($action eq "")                     {  print "job-other 42: wrong action - must be not empty\n";      exit; }
if (!("run,cancel,delete"=~/$action/)) {  print "job-other 43: wrong action - must be run: $action \n";  exit; }

#cancel or delete
if ($action eq "cancel") { &my_cancel; $info="cancel job"; &my_log; exit; }
if ($action eq "delete") { &my_delete; $info="delete job"; exit; }

#debug via CLI call
# print "job-snap 54: run id=$id par $par\n";
sleep(5);  # small delay to replication snaps

# delay to start replications at the same time first
sleep(8);

$error=&my_snap;

&my_log;
exit;


 ## delete job
 sub my_delete {
 ##############
 my ($r);
      if ($id ne "") {
        $r=`sudo rm -f /var/web-gui/_log/jobs/$id.*`;
        print "##### job-72: delete $id ######  $r\n";
      } else {
        print "job-74 delete, no id !! ########### \n";
      }
 }
#############

## cancel job
sub my_cancel  {
###############
   if (-f "/tmp/nappit/$id.progress")  {  unlink("/tmp/nappit/$id.progress"); }
   if (-f "/tmp/nappit/$id.cancel")    {  unlink("/tmp/nappit/$id.cancel"); }

   #optionally cancel this job (longrun with md5)
   &my_logid('cancel','83 cancel requested');
   my $r; #=`sudo ps axw | grep job-snap | grep $id`;
   if ($r ne "") {
       $r=~s/^ +//; $r=~s/ .*//;   # keep psid  and kill
       `kill -9 $r >/dev/null 2>&1`;
   }

 }

############# hold snaps for n days
sub my_hold {
############

  my $s=$_[0];

  # check hold (days)
  $t=$id."_hold";
  $t=$current{$t};         # empty or 0..n
  $t=~s/\D//g;

  my $n=$s;
  $n=~s/.*\_//;

  #debug
  #&my_logid("hold settings","hold $t days:");

  if ($t > 0) {

      my $days=`$sys{'sbinpath'}zfs list -H -o creation $s`;  # ex Sun Sep 15  0:05 2013
      $days=~s/ +/ /g;

      #debug
      #&my_logid("snap","$s creation: $days:");

      my @w=split(/ /,$days);

      #month in w[1], day in w[2], year in w[4]
      $w[1]=~s/Jan/01/; $w[1]=~s/Feb/02/; $w[1]=~s/Mar/03/; $w[1]=~s/Apr/04/; $w[1]=~s/May/05/; $w[1]=~s/Jun/06/;
      $w[1]=~s/Jul/07/; $w[1]=~s/Aug/08/; $w[1]=~s/Sep/09/; $w[1]=~s/Oct/10/; $w[1]=~s/Nov/11/; $w[1]=~s/Dec/12/;

      $days=&date2diff("$w[2].$w[1].$w[4]","$sys{'mday'}.$sys{'mon'}.$sys{'year'}");

      #debug
      # &my_logid("snapdiff","$s age: $days, hold $t days");

      if ($days < $t) {            # age < hold

          # log
          &my_logid("hold nr $n","hold $t days: match rule: snap-age $days days");
          return('1');                      # skip delete
      }
  }

  # ok, no extra hold
  return('0');


}

############  execute prior ZFS snap
sub my_pre {
############
   my ($r,$t,@l,$l,$c);
   &my_logid("execute $id.pre");
   if (-f "/var/web-gui/_log/jobs/$id.pre") {
        $r=`cat /var/web-gui/_log/jobs/$id.pre`;
        @l=split(/\n/,$r);
        foreach $l (@l) {
             if ($l=~/^#/) { next; }     # comment
             $l=~s/=date_and_time=/$current{'timestamp'}/;
             $c="";
             if ($l=~/\#/) {
                $c=$l; $c=~s/.*\#/ #/;
             }
             $l=~s/;.*//; $l=~s/#.*//;
             $l.=' 2>&1';  # incl. error
             $l.="$c";     # comment

print "sudo sh $l\n";
             &my_log_last("exe: $l");
             $t=`sudo $l`;

print "$t\n";

            $t=~s/\n/<b>/gs;
             if ($t=~/failed|unable/is) {
                 &my_logid("error;$l;$t");
             }
             &my_log_last("result: $t");
        }
   }

}

############  execute post ZFS snap
sub my_post {
############
   my ($r,$t,@l,$l,$c);
   &my_logid("execute $id.post");
   if (-f "/var/web-gui/_log/jobs/$id.post") {
        $r=`cat /var/web-gui/_log/jobs/$id.post`;
        @l=split(/\n/,$r);
        foreach $l (@l) {
             if ($l=~/^#/) { next; }     # comment
             $c="";
             if ($l=~/\#/) {
                $c=$l; $c=~s/.*\#/ #/;
             }
             $l=~s/;.*//; $l=~s/#.*//; $l.=' 2>&1'; # incl. error
             $l.="$c";     # comment

             &my_log_last("exe: $l");
             $t=`sudo $l`;

             $t=~s/\n/<b>/gs;
             if ($t=~/failed|unable/is) {
                 &my_logid("error;$l;$t");
             }
             &my_log_last("result: $t");
        }
   }
}

#############      del snaps
sub my_snap  {
##############

  my ($case,$r,$time,$snaps,@snaps,$i,$s,$name,$old,$keep,@w,$delzero,$pos,$pool,$snapname,$f,$rec,$q,$hold,$count,$count_k,$count_z,$hold,$used);
  $old="";

  &getdate;
  $time=$sys{'date_all'};                                         # yyyy.mm.dd.mm.ss (year.month.day.min.sec)

  # read parameters for current job
  $t=$id."_keep";
  $keep="$current{$t}";
  $keep=~s/\D//g;
                                          # 1..n
  if (!($keep >=1)) { $keep=1; }

  $t=$id."_hold";
  $hold="$current{$t}";
  $hold=~s/\D//g;

  $t=$id."_delzero";
  $delzero="$current{$t}";                                        # yes, delete size=0

  print "keep=$keep\nhold=$hold\ndelzero=$delzero\n";

#  if ($delzero ne "") {
#     $keep++;                                                     # keep = amount of snaps + one zero sized
#  }

  $t=$id."_name";
  $name=$current{$t};

  $t=$id."_pool";
  $pool=$current{$t};                                             # ZFS not pool ex /tank/data

  #recursive snaps
  $rec=" -r";                                                     # default
  $t=$id."_disable-rec";
  if ($current{$t}=~/yes/i) {
     $rec="";
  }

  # print "name=$name, delzero=$delzero, keep=$keep, pool=$pool\n";
  if (($name eq "") || ($pool eq ""))  { $error="pool or jobname missing"; return; }

  # zfs  not mounted/zfs not available
  #$t=&exe("zfs list -H -o name | grep $pool 2> /dev/null");
  $q=quotemeta($pool);
  if (!($t=~/$q/)) {
    &my_logid("ZFS filesystem $pool not mounted/available");
    print "zfs=$pool not available\n"; return;
  }


  #execute a .pre
  $t=$id."_prejob";
  if ($current{$t} eq "yes") {
    &my_pre;
  }

  # ESXi snaps
  $t=$id."_esxi_ip";
  if ($current{$t} =~/\d/) {
    $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl create_$id`;   # create ESXi snaps via soap
  }

  # create new snapshot
  # skip in case summer -> wintertime (snap at time already created)
      #$r=&exe("zfs list -t snapshot | grep $pool\@$name\-$id\_$time");
      if ($r ne "") { return(); }

      $snapname="$pool\@$name\-$id\_$time";
      $info="zfs snapshot$rec $snapname";

      print "create: zfs snapshot$rec $snapname\n";
      $error=&exe("zfs snapshot$rec $snapname");                  # ex tank@weekly-jobid_date
      if ($error ne "") { $error=" line 116 $error"; &my_log; }   # already created with exact timestamp

  #delete snaps older max with names pool..@name_id               #
      #$snaps=&zfslib_list_datasnaps('f');                         # commalist of snaps

      #$snaps=&exe("zfs list -H -r -t snapshot -o name,used,creation -S creation");      # sort creation time not needed, date in name
      $snaps=&exe("zfs list -H -r -t snapshot -o name,used,creation");

      chomp($snaps);
      $snaps=~s/\n/,/gs;

      @snaps=split(/,/,$snaps);                                    # sort is newest first

      $i=0;
      $count=0;
      $count_k=0;
      $count_z=0;

      @snaps=sort(@snaps);
      @snaps=reverse(@snaps);           # newest first

      foreach $s (@snaps) {

          @w=();
          @w=split(/\s+/,$s);

          if (!($w[0]=~/\@/))     { next; }                              # not a snap
          if (!($w[0]=~/$id/))    { next; }                              # not my id
          if ($w[0]=~/\_repli\_/) { next; }                              # do not touch replications

          # ex

          # daten/vm@daily-1513546920_2019.07.27.11.03.48          # name
          # 0                                                      # used
          # Sat Jul 27 12:55 2019                                  # cration

         $name=$w[0];

         # we need only names of snaps without date (ex tank@15min-1357655234_2013.05.27.09.00.02 -> tank@15min-1357655234)
         $pos=rindex($name,"_");                                    # position of last "_"
         $name=substr($name,0,$pos);                                # snapname without timestamp

         if ($name ne $old) { $i = 0; } else { $i++; }              # new pool or snapshot, reset count

 print "\nscheck $i $w[0]  used: $w[1]\n old=$old\n new=$name\n";

         if ($delzero eq "yes")  {
            $used=$w[1];
         } else {
            $used="1";
         }

         # always care about days and number
         if ($i >= $keep) {                                         # more snaps than wanted
              $r=&my_hold($w[0]);                                   # respect hold 0/1?
              if ($r eq "0") {
                 print " kh $i: zfs destroy $w[0] \n";
                 &exe("zfs destroy \"$w[0]\"");                     #  do not hold and > keep
                 $count++;
                 $count_k++;
              }
         }

         # delete snaps with size=0 in case 1 and 2
         if ($used eq "0")  {                             # delete all size=0
            if ($i>0)  {  # always keep at least newest even zero snap
              # if delzero is activated later there may be some former 0 snaps now <>0
              # so re verify
              $t=&exe("zfs list -Ho used \"$w[0]\"");
              chomp($t);
              if ($t==0) {
                print " zero $i: zfs destroy $w[0] $w[1] \n";
                &exe("zfs destroy \"$w[0]\"");
                $count++;
                $count_z++;
              }
            }
         }
                                                                   # remember last name
         $old=$name;

      } #end foreach snaps


      #post
      #execute a .post
      $t=$id."_prejob";
      if ($current{$t} eq "yes") {
        &my_post;
      }

      # ESXi snaps
      $t=$id."_esxi_ip";
      if ($current{$t} =~/\d/) {
        $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl remove_$id`;   # remove ESXi snaps, can happen to fail on some snaps due "too busy"
        $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl remove_$id`;   # check/redo, second run to be sure all are removed"
      }

      &my_logid("destroyed $count snaps k/h=$count_k zero=$count_z");
      print "destroyed $count snaps k/h=$count_k zero=$count_z\n";
      if (-f "/tmp/nappit/zfs_last.zfs") { `sudo rm /tmp/nappit/*.zfs`;  }

      #md5 hash via 'digest -a md5 -v file' in a folder or ZFS /pool/validate.
      $t=$id."_md5";
      if ($current{$t} eq "yes") {                                  # yes/no

           &task_gettime;                                           # $current{'timestamp'}, $current{'logtime'}
           $t=$pool; $t=~s/\/.*//;  &getp;                          # get real poolname
           my $folder="/$t/validate";
           if (!(-d $folder))  {  mkdir ($folder);  }

           my $file="$name\_$time.log";
           $file=~s/\//\~/g;
           $file="$folder/$file";

           #debug only via CLI call
           # print "logfile:$file\n";
           if (!($cfg{'ver'}=~/c|m/)) { return; }

           open (PF, ">>$file");
                flock(PF, LOCK_EX);
                print PF "Create md5 /usr/bin/digest -a md5 -v file at $current{'timestamp'} for $snapname\n\n";
           close (PF);

           $r=`find \'/$pool\' -type d`;                           # find folders in zfs
           chomp($r);
           my @fo=split(/\n/,$r);

           my @fi=();
           my ($p,$fi);
           foreach my $f (@fo) {
                 if ($f=~/\/\./) { next; }                        # not on hidden folders

                 #read files
                 @fi=();
                 $p=$f;                                           # current folder
                 $p=~s/^\'|\'$//g;                                # remask
                 $p=~s/\/$//;

                 $p=&myfn2fn($p);                                   # allow ' in name

                 $r=`/usr/bin/ls -1AlU --time-style=full-iso \'$p\'`;

                 # debug only
                 #&log_reset("read files: ls -1AlU --time-style=full-iso \'$p\'\n$r\n\n");

                 # one line per file like
                 # 'total 270070'
                 # 'drwxrwxrwx+  2 root     root           2 2013-02-19 10:15:23.152443377 +0100 Neuer Ordner'
                 # '-rwxrwxrwx+  1 root     root           0 2013-02-19 09:29:11.134576933 +0100 b.txt'


                 @fi=split(/\n/,$r);                         # ", \n"  or ", " is delimiter
                 $r="";
                 foreach $fi (@fi) {

                          #debug
                          # &log_reset("debug: $fi\n");                       # only for debugging

                          chomp($fi);
                          if ($fi=~/^\./)       { next; }                   # not on hidden files
                          if ($fi=~/^d/)        { next; }                   # not on folders
                          if (!($fi=~/^\-/))    { next; }                   # not a file like total

                          $fi=~s/.*\d+ \+\d+ //;                            # keeps filename

                          $fi="$p/$fi";

                          $fi=&myfn2fn($fi);                                  # allow ' in name
                          $fi="\'$fi\'";

                          #get md5
                          $r=`/usr/bin/ls -l $fi`;
                          chomp($r);
                          $r.=" ";
                          $t=`/usr/bin/digest -a md5 -v $fi`;
                          chomp($t);
                          $t=~s/.*=/md5 =/;

                          #log

                           open (PF, ">>$file");
                               flock(PF, LOCK_EX);
                                   print PF "$t $r\n";
                           close (PF);

                           #debug only
                           # print "$r $t\n";

                 }  # end files
           } # end folder
           &my_logid("added md5; $file");

           $error="";
      } # end md5

 }



###########  ## log last run in jobfile /_log/jobs/id...job
sub my_log {
###########

  # print "job-repli 548 - call log id=$id action=$action\n";

  # line 1: lastrun = ok or err
  # line 2: time of last run
  # line 3+: info

  &task_gettime;                                             # $current{'timestamp'}, $current{'logtime'}

  my (@v,$v,$n,$l,$f1,$f2,@data);


  # ha config an pool missing?
  if (-f "/var/web-gui/_log/ha-plugin.txt") {
     $n=&exe("cat /var/web-gui/_log/ha-plugin.txt");
     $n=~s/\n.*//s; chomp ($n);
     if ($n=~/\w/) {
       if (!(-d "/$n")) {
          &my_log_last("HA pool $n not available");
          exit;
       }
     }
  }



  $l=$error;
  if ($l eq "") { $l="ok"; } else { $l="error"; $info="$info $error"; }

  $l=~s/\n/ /gs;
  $info=~s/\n.*//s;                                        # first line
  $info=~s/ +/ /g;

 #stage 1: log general jobs
 ###########################
  my $log="/var/web-gui/_log/jobs/jobs.log";
  my $lock="/var/web-gui/_log/jobs/jobs.lock";
  open (X,"<$lock");
  flock(X, LOCK_EX);
          open (F,"<$log");
              flock(F, LOCK_EX);
              @data=<F>;
          close (F);

        #unshift   $l=ok/error
        unshift(@data,"$l; snap $ref $id; time: $current{'timestamp'}; info: $info\n");

        if (@data > 50) {
                   $t=@data-50;
                   splice(@data,50,$t);
        }

       $t=join("",@data);
       open (PF, ">$log");
          flock(PF, LOCK_EX);
          print PF "$t";
       close (PF);
 close (X);

 if ($info=~/^delete /) { exit; }


 #stage 2 update job.par  $current{'logtime'};
  $log="/var/web-gui/_log/jobs/$id.log";

  $t="$id"."_last";                                              # update last
  $current{$t}=$current{'logtime'};
  &task_update_par($id);                                         # update last


  # stage 3 id.log newest at top
  ###################
  &my_logid("$l","info: $info");

  # update .par
  &task_gettime;                                                # $current{'timestamp'}, $current{'logtime'}
  $t="$id"."_last";                                             # update last
  $current{$t}=$current{'logtime'};
  &task_update_par($id);

 exit;  # log always last action

}

########### log detailed info for last job (klick on Date of last exec)
sub my_log_last {
##########
   &task_gettime;
   my $t=$_[0];

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);

                  my $l="$p0[3] line $p0[2]";             $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $l.="<br> <- $p1[3] line $p1[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $l.="<br> <- $p2[3] line $p2[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $l.="<br> <- $p3[3] line $p3[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $l.="<br> <- $p4[3] line $p4[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   $l=~s/main::/\&/gs;
   $l=~s/br> +/br> /g;
   $l=~s/\s+/ /g;

   $l=~s/\n/<br>/g;
   $t=~s/\n/<br>/gs;


   my $log="/var/web-gui/_log/jobs/$id.last";
   open (PF, ">>$log");
     flock(PF, LOCK_EX);
     print PF "$current{'timestamp'}\t$l\t$t\n";
   close (PF);
}


##############  log to id.log
sub my_logid {
#############
my $err =$_[0];
my $data=$_[1];
$data=~s/ +/ /g;
my @data;
&task_gettime;

my $log="/var/web-gui/_log/jobs/$id.log";
  open (F,"<$log");
        flock(F, LOCK_EX);
        @data=<F>;
  close (F);

  #unshift   $l=ok/error
  unshift(@data,"$err; time: $current{'timestamp'}; $data\n");

  if (@data > 50) {
             $t=@data-50;
             splice(@data,50,$t);                 # delete from 50+
  }

 $t=join("",@data);
 open (PF, ">$log");
    flock(PF, LOCK_EX);
    print PF "$t";
 close (PF);
}

sub myfn2fn {           # hide:  parse ' and " in filenames   quoting in unix: don't  -> 'don'\''t' or "don't"
    my $t=$_[0];
    $t=~s/^\'//;
    $t=~s/\'$//;               #

    $t=~s/\'\\\'\'/\'/g;       # do not double  '\'' -> '
    $t=~s/\'/\'\\\'\'/g;       # ' -> '\''

    chomp($t);
    return ($t);
}

##
1;